#!/usr/bin/perl
###############################################################################
# $Id$
#
# Copyright(c) 2006-2014 ThreatSTOP, Inc. All Rights Reserved
# 
# NOTICE:  All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and may be covered by U.S. and Foreign Patents, patents
# in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# ThreatSTOP, Inc.
#
#
# ThreatSTOP main function module
################################################################################

package ThreatStop;
use Exporter;

use strict;
use warnings;
use Scalar::Util qw(openhandle);
use Cwd qw(abs_path);
use Fcntl qw(:flock);
use File::Basename;
use FindBin;
use Data::Dumper;
use MIME::Base64 qw(encode_base64 decode_base64);
FindBin->again;
###############################################################
################## MODULE DEFENITIONS PART ####################
###############################################################
our $VERSION   = 1.00;
our @ISA       = qw(Exporter);
our @EXPORT    = qw(getCurrentTime logInit printMsg logMsg setLock setUnLock ipv4_cidr2msk checkDNSaccess updateDNSlist calculateElapsedTime loadConfig intHandler osPlatformDetecter loadConfFileData setWorkingDir setAlarm osDestributionDetector $DNS_LIST $DEBUG_LEVEL $LOGFILE $LOG_LEVEL $dns_server $port);
#$DNS_PROTO $DNS_PROTO_ALT $DNS_DOMAIN_BASE $DNS_DOMAIN_ALT $DNS_SERVERS_ALT $DNS_SERVERS_LCL $DNS_MAX_TRIES $DNS_LIST_NEW);

our $CFG_DATA;
our $CONFIG_PATH = "/etc/threatstop";
our $CONFIGFILE  = "$CONFIG_PATH/threatstop.conf" if ( ! defined ($CONFIGFILE));
our $LOGFILE     = $FindBin::Bin . "/tsoutput.log" if ( ! defined $LOGFILE); #Define default log
our $LOCKFILE    = $0.".lck" if ( ! defined ($LOCKFILE));
our $LOGINIT     = 0;
our $DEBUG_LEVEL = 5 if ( ! defined $DEBUG_LEVEL);                        #Define verbose level 5-nothing
our $LOG_LEVEL   = 6 if ( ! defined $LOG_LEVEL);                          #Define verbose level 5-nothing
our ($FD, $FDL, $LFD);
our $START_EPOCH = time();
our $END_EPOCH;

our $DNS_LIST = "64.87.26.147 24.249.204.58";
our $dns_server;                                                          # "64.87.26.147 24.249.204.58";
our $DNS_PROTO       = '+tcp';					          # Protocol to use - TCP
our $DNS_PROTO_ALT   = '+notcp';                                          # Protocol to use - UDP
our $DNS_PORT_BASE   = 53;                                                # Port to use (default 53)
our $DNS_PORT_ALT    = 5353;                                              # Port to use (alternate 5353)
our $DNS_DOMAIN_BASE = 'dns-servers.threatstop.local';                    # Server name to resolve 
our $DNS_DOMAIN_ALT  = 'dns-servers.threatstop.com';                      # Alternative server name to resolve
our $DNS_SERVERS_ALT = "8.8.8.8 8.8.4.4 208.67.220.220 208.67.222.222";   # Alternate servers ip's to quiery
our $DNS_SERVERS_LCL = '127.0.0.1';                                       # Local DNS serves will be stored here
our $DNS_MAX_TRIES   = 3;                                                 # Number of tries to update DNS list
our $DNS_LIST_NEW;
our $DNS_PORT        = 53;                                                # Actually we read this param from conf file

# Setup signal handler for possible user interupts
$SIG{'INT'}  = 'intHandler';
$SIG{'ABRT'} = 'intHandler';
$SIG{'QUIT'} = 'intHandler';
$SIG{'STOP'} = 'intHandler';

###############################################################
################# FUNCTIONS DEFENITIONS PART ##################
###############################################################

############################################
########## PRIVATE FUNCTIONS ###############
############################################

################################################################

##################################################
# intHandler : Used to trap exit signals and per #
#              form lockfile cleanup             #
##################################################
sub intHandler() {
        if ( defined $FDL ) {
                close($FDL);
                unlink($LOCKFILE);
        }

        print "\n";
        exit(0);
}

##################################################
# loadConfig : Loads additional parameters from  #
#              conf file                         #
##################################################
sub loadConfig() {

        &printMsg("INFO","Loading configuration file ...");

        eval {
                open(FD,"<$CONFIGFILE") or die "Failed to open config file [$CONFIGFILE] with error [$!]";
        };

        if ( $@ ) {
                &printMsg("ERROR","$@");
                return 1;
        }

        while (<FD>) {
                chomp;                  # no newline
                s/#.*//;                # no comments
                s/^\s+//;               # no leading white
                s/\s+$//;               # no trailing white
                s/'//g;                 # no ticks
                next unless length;     # anything left?
                my ($var, $value) = split(/\s*=\s*/, $_, 2);
                $CFG_DATA->{$var} = $value;
        }

	close FD;
    mpw($CONFIGFILE,$CFG_DATA);
	return 0;
}

#mangle/unmangle config must be called right after loading in config values. 
sub mpw {
   my $conf = shift; 
   my $confref = shift;
   foreach my $i (qw(password enable_pw)) {
       if(defined($confref->{$i})) {
           my $tmp = $confref->{$i};
           $tmp =~ s/\"//g;
           if($tmp =~ /^\$\$.*\$\$$/) {
               $tmp =~ s/\$//g;
               $tmp =~ s/!/=/g;
               $tmp = decode_base64($tmp);
#               print "$i encoded\n";
#               print "$i: $tmp\n";
               $confref->{$i} = $tmp;
           }
           else {
#               print "$i plaintext\n";
#               print "$i: $confref->{$i}\n";
               my $tmp = $confref->{$i};
               $tmp =~ s/\"//g;
               $tmp = encode_base64($tmp);
               chomp($tmp);
               $tmp =~ s/=/!/g;
               $tmp = '"' . '$$' . $tmp . '$$' . '"';
               rewriteConfKey($conf,$i,$tmp);
           }
           
       }
   }
}

#Re-write the given key/value pair in the specified config file.
sub rewriteConfKey() {
   my $conf = shift;
   my $key = shift;
   my $value = shift;
   my $in;
   my $out;
   my @data;
   eval {
       open($in,'<',$conf);
       @data = <$in>;
       close($in);
       open($out,'>',$conf);
   };
   if ( $@ ) {
       &printMsg("ERROR",$@);
       return 1;
   }
   foreach my $line (@data) {
       if($line =~ /^$key=/) {
           print $out "$key=$value\n";
           &printMsg("DEBUG","rewrote $key = $value in $conf");
       } else {
           print $out $line;
       }
   }
   close($out);
   return 0;

}
#####################################################################
sub readLocalBoxNameserversLst() {

        my $servers_lst = shift;
        my $cmd = "cat /etc/resolv.conf | grep nameserver | /usr/bin/awk '{print \$2}'";
        my $rc = qx($cmd);

        &printMsg("INFO","Reading local box configuration");
        $rc =~ tr/\n/ /;
        $$servers_lst = $rc;

        return 0;
}

#####################################################
sub updateDNSConfFile() {

	my $in;
	my $out;
	my $isPort=0;
	my @cnt;
	
	eval {
		open($in,'<',$CONFIGFILE) or die "Failed openig [$CONFIGFILE] file for reading";
	};

	if ( $@ ) {
		&printMsg("ERROR",$@);
		return 1;
	}

	@cnt=<$in>;
	close $in;

	eval {
		open($out,'>',$CONFIGFILE) or die "Failed openig [$CONFIGFILE] file for writing";
	};

	if ( $@ ) {
		&printMsg("ERROR",$@);
		return 1;
	}

	foreach my $line ( @cnt ) {
		#We have found port setting.
		$isPort = 1 if ( $line =~ /port=/);                      # We have found port setting
		$line =~ s/port=.*/port=$DNS_PORT/gi;                    # Update port setting
		$line =~ s/dns_server=.*/dns_server=\"$dns_server\"/gi;  # Update DNS list setting

		print $out $line;
	}

	if ( $isPort == 0 ) {
		#We do not encounted port setting while updating conf file
		#than we need to add one
		print $out "\n";
		print $out "port=$DNS_PORT\n";
	}
		
	close $out;
        return 0;
}

##################################################################
# compareDNSlists :                                              #
#                   $dns_list - list with DNS ips                #
#                   $dns_res  - hash with results of DNS quering #
##################################################################
sub compareDNSlists() {

        my $dns_list = shift;
        my $dns_res  = shift;

        &printMsg("INFO","Comparing DNS lists");

        #For each DNS found in both list increase count. Than see what DNS servers
        #counters had been left intact
        foreach my $dns ( split(/ /,$dns_list) ) {

                #This entry is in two lists. Increase hash counter  
                if ( exists $dns_res->{$dns} ) {
                        $dns_res->{'dnslist'}->{$dns} = 2;
                        next;
                }

                #This entry does not appear in second list. Means two lists are 
                #different. No need to continue. Use new list for future checks
                if ( ! defined $dns_res->{$dns} ) {
                        &printMsg("DEBUG","Old and new lists are differnet");
                        return 1;
                }
        }

        #Now see if there are entries in dns_list that were not updated by previous loop
        #meaning there are entries in new list which are not in original list. If found
        #no need to continue. Use new list for future checks 
        foreach my $dns ( keys %{$dns_res->{'dnslist'}} ) {
                if ( $dns_res->{'dnslist'}->{$dns} == 1 ) {
                        &printMsg("DEBUG","DNS lists are different for [$dns]");
                        return 1;
                }
        }

        #Both lists are similar
        &printMsg("DEBUG","Both lists are similar");
        return 0;
}

########################################################################################
# getDNSserversList :                                                                  #
#                    $dig             - dig tool                                       #
#                    $dns_to_ask      - DNS server ip to query                         #
#                    $dns_to_resolve  - DNS domain name to resolve                     #
#                    $dns_protocol    - Net protocol to use TCP/UDP                    #
#                    $dns_query_type  - Quiery type ptr/apl/a (default not set)        #
#                    $dns_status      - ref to hash where query results will be stored #
########################################################################################
sub getDNSserversList() {

        my $dig            = shift;
        my $dns_to_ask     = shift;
        my $dns_to_resolve = shift;
        my $dns_protocol   = shift;
	my $dns_port       = shift;
        my $dns_query_type = shift;
        my $dns_status     = shift;
        my $rc;
        my $rv;
        my @q_line;
        my @rc_data;
        my $cmd;
        my $idx;

        # Quering DNS server
        $cmd = "$dig $dns_protocol $dns_query_type -p $dns_port \@$dns_to_ask $dns_to_resolve 2>&1";
        $rc = qx($cmd);
        $rv = $?;

        &printMsg("DEBUG","$cmd");

        #If dig command failed. Server did not respond go to the next one
        if ( $rv != 0 ) {
                &printMsg('WARN',"DNS server [$dns_to_ask] did not respnod");
                return 0;
        }

        #################################################
        #Parse command output and extract relevant info # 
        #      - response state                         #
        #      - response time                          #
        #      - DNS servers list                       #
        #################################################
        @rc_data = split(/\n/,$rc);

        for($idx=0; $idx<$#rc_data; $idx++) {

                # Response status
                if ( $rc_data[$idx] =~ /;; ->>HEADER<<- opcode: QUERY, status:/ ) {

                        my $status = (split(/ /,$rc_data[$idx]))[5];

                        #If status is not NOERROR. Do not use this server. This lines is analysed
                        #prior to other data. Simply return is suficient to not set settings for
                        #this server
                        if ( $status !~ /(NOERROR)/ ) {
                                &printMsg('WARN',"DNS server [$dns_to_ask] respond with bad query status");
                                return 0;
                        }

                        #Query status ok, mark server as good to use 
                        $dns_status->{$dns_to_ask}->{'res_status'} = $1;

                        next;
                }

                # Response time
                if ( $rc_data[$idx] =~ /;; Query time:/ ) {
                        $dns_status->{$dns_to_ask}->{'res_time'} = (split(/ /,$rc_data[$idx]))[3];
			#print "time - $dns_to_ask - $dns_status->{$dns_to_ask}->{'res_time'}\n";
                        next;
                }

                # DNS servers list
                if ( $rc_data[$idx] =~ /;; ANSWER SECTION:/ ) {

                        $idx++; # Take next line
                        while ( $rc_data[$idx] !~ /;;/ ) {
                                if ( $rc_data[$idx] =~ /IN\tA/ ) {
					#Split response line. Take DNS ip. Also set for this DNS ip
					#fake response time and return status. We will update it
					#later
                                        my $dns_ip = (split(/[\t|\s]/,$rc_data[$idx]))[4];
					if ( ! defined $dns_status->{'dnslist'}->{$dns_ip} ) {
	                                        $dns_status->{'dnslist'}->{$dns_ip}    = 1;
						$dns_status->{$dns_ip}->{'res_time'}   = 0;					
						$dns_status->{$dns_ip}->{'res_status'} = 'UNKNOWN';
					}
                                }
                                $idx++;
                        }
                        #Go back one line it some queries it can me the beginning of QUERY
                        #section and we defenetly want to parse it
                        $idx--;
                        next;
                }
        }

	#Now make sure that we do not take with us servers that are not in DNS list. This can happened
	#if we switched to global servers list. WE do not want them only resolve through them
	foreach my $ips ( keys %$dns_status ) {

		if ( $ips eq 'dnslist' ) {
			#Scip this key. It holds DNS list
			next;
		}

		if ( ! defined $dns_status->{'dnslist'}->{$ips} ) {
			&printMsg("DEBUG","Removing server [$ips] from the list ");
			delete $dns_status->{$ips};
		}
	}

	#print Dumper($dns_status);

        return 0;
}

#################################################################################
# setNewDNSserversList : checks DNS servers response time and put the quickest  #
#                        to response server at the begining of the list         # 
#################################################################################
sub setNewDNSserversList() {

        my $dns_res     = shift;
        my $dns_res_lst = shift;
        my $dns_tmp;
        my $res_time = 0;
        my $res_server;
        my $cmd;
        my $rv;

        &printMsg("DEBUG","Determining the quickest to respond DNS server");

        #Look for server with shortest response time
        foreach my $dns ( keys %{$dns_res->{'dnslist'}} ) {

                my $srv_res_time = $dns_res->{$dns}->{'res_time'};

		#If this server failed to respond than there is no response time
		#setting. Set it anyway to 10000 to make sure we do not choose this server
		$srv_res_time = 10000 if ( ! defined $srv_res_time );


                if ( $res_time == 0 ) {
                        #This is the first item in the list
                        $res_time = $srv_res_time;
                        $res_server = $dns;
                        next;
                }

                if ( $res_time > $srv_res_time ) {
                        #$res_time = $srv_res_time if ( defined $srv_res_time);
                        $res_time = $srv_res_time;
                        $res_server = $dns;
                        next;
                }
        }

        #Update the $dns_list variable with new servers set. Put the quickest server at the 
        #begining of the list
        $dns_tmp = $res_server;

        foreach my $dns ( keys %{$dns_res->{'dnslist'}} ) {
                next if ( $dns eq $res_server ); #Scip we already have put this server in the list 
                $dns_tmp .= " $dns";
        }

        $$dns_res_lst = $dns_tmp;

        return 0;
}

###############################################################################
# updateDNSlistsDo :                                                          #
#                    $dns_lists   - list of DNS servers ip's to run check on  #
#                    $dns_domain  - server full name to resolve               #
#                    $dns_proto   - protocol to use tcp/udp                   #
#                    $dns_res_lst - DNS list created by our quieries          #
#                    $dns_res     - hash with query parsed results            #
###############################################################################
sub updateDNSlistsDo() {

	my $dig_tool    = shift;
        my $dns_list    = shift;
        my $dns_domain  = shift;
        my $dns_proto   = shift;
	my $dns_port    = shift;
        my $dns_res_lst = shift;

        my $dns_res     = {};
        my $dns_cmp;

        &printMsg("INFO","Testing [$dns_list] servers");

        #foreach server in the DNS list query it and extract returned list of DNS servers.
        #store it in hash for future comparison with original DNS list
        foreach my $dns ( split(/ /,$dns_list)) {
                &printMsg("DEBUG","Querying [$dns] server on port [$dns_port] protocol [$dns_proto] domain [$dns_domain]");
                return 1 if (&getDNSserversList($dig_tool,$dns,$dns_domain,$dns_proto,$dns_port,'',$dns_res));
        }

        #If no entries were created in $dns_res this means all calls to currently defined DNS servers
        #failed go back and re-rty with differnet options
        if ( ! keys %$dns_res ) {

                &printMsg("WARN","All servers for domain failed to resolve...trying to fall back...");

                return 2 if ( $dns_domain eq $DNS_DOMAIN_BASE && 
                               $dns_port   eq $DNS_PORT_BASE   );     #we testes for base domain

		return 3 if ( $dns_domain  eq $DNS_DOMAIN_BASE &&
                               $dns_port   eq $DNS_PORT_ALT   );      #we tried for base domain on alternate port 

                return 4 if ( $dns_domain eq $DNS_DOMAIN_ALT  &&
                               $dns_list   eq $dns_server         );  #we testes for alt domain on base port

                return 5 if ( $dns_domain eq $DNS_DOMAIN_ALT  &&
                               $dns_list   eq $DNS_SERVERS_ALT    );  #we tested for global serves on base port

                return 6 if ( $dns_domain eq $DNS_DOMAIN_ALT  &&
                               $dns_list   eq $DNS_SERVERS_LCL &&
                               $dns_proto  eq $DNS_PROTO          );  #we tested for local defined servers on base port

                return 7 if ( $dns_proto  eq $DNS_PROTO_ALT &&
                              $dns_list   eq $DNS_SERVERS_LCL  );    #we tested for UDP protocol on base port
		
                return 1;                                             #Unrecognized case
        }

        #Find the quickes to reponse DNS server put it at the beggining of the list. 
        return 1 if ( &setNewDNSserversList($dns_res,$dns_res_lst) != 0 );

        #Ok, we have results of DNS query DNS resulting list. Compare it to list we start
        #quering with. Hopefully they will be completly similar
	return 8 if ( &compareDNSlists($dns_list,$dns_res) == 0 );

        #The DNS servers respond ok, but returned DNS list is differnet than the original one
        #go back and re-run operation with new list settings
        return 9;
}

############################################
########### PUBLIC FUNCTIONS ###############
############################################

##################################################
# getCurrentTime : Retruns curretn time stamp    #
##################################################
sub getCurrentTime() {

        my $tr = shift;
        my @td = localtime(time);
        my $ts = sprintf("%02d/%02d/%04d %02d:%02d:%02d",$td[3],$td[4]+1,$td[5]+1900,$td[2],$td[1],$td[0]);
        
        $$tr = $ts;

        return 0;
}

##################################################
#logInit : initialize logging mechanism          #
##################################################
sub logInit() {

	my $logfile = shift;

	#If logfile supplied use it instead of default
	$LOGFILE = $logfile if ( defined $logfile ); 
	$LOGINIT = 1;

    #sometimes the script gets run as root by hand and we 
    #wind up with an unwritable log file this check catches this case
    #and attempts to unlink the file before we open for write.
    if( -f $LOGFILE) {
        if ( ! -w $LOGFILE ) {
            &printMsg("WARN","Couldn't open $LOGFILE for writing attempting to unlink");
            if(unlink($LOGFILE)) {
                &printMsg("WARN","Successfully unlinked $LOGFILE");
            }
        }
    }
        eval {
            open(LFD, '>' . $LOGFILE) or die $!;
        };

	if ( $@ ) {
		&printMsg("WARN","Can not initialize logging ... Proceeding without ...");
		return 0;
	}

	return 0;
}

##################################################
#logMsg : prints info into log file              #
##################################################
sub logMsg() {
	
	my $log_type = shift;
        my $log_body = join(' ',@_);

        my $log_out = ( $log_type eq 'ERROR'                    ) ? "[ERROR] : $log_body"
                    : ( $log_type eq 'ALERT' && $LOG_LEVEL >= 1 ) ? "[ALERT] : $log_body"
                    : ( $log_type eq 'WARN'  && $LOG_LEVEL >= 2 ) ? "[WARN ] : $log_body"
                    : ( $log_type eq 'INFO'  && $LOG_LEVEL >= 3 ) ? "[INFO ] : $log_body"
                    : ( $log_type eq 'DISPL' && $LOG_LEVEL >= 4 ) ? $log_body
                    : ( $log_type eq 'NOTE'  && $LOG_LEVEL >= 5 ) ? "[NOTE ] : $log_body"
                    : ( $log_type eq 'DEBUG' && $LOG_LEVEL >= 6 ) ? "\t[DEBUG] : $log_body"
                    :                                               undef;

	if ( defined $log_out ) {
		print LFD $log_out . "\n" if ( $LOGINIT == 1 );
	}

	return 0;
}

##################################################
# printMsg : Prints/logs different level output  #
##################################################
sub printMsg() {

	my $msg_type = shift;
        my $msg_body = join(' ',@_);

        my $msg_out = ( $msg_type eq 'ERROR'                    ) ? "[ERROR] : $msg_body"
                    : ( $msg_type eq 'ALERT' && $LOG_LEVEL >= 1 ) ? "[ALERT] : $msg_body"
                    : ( $msg_type eq 'WARN'  && $LOG_LEVEL >= 2 ) ? "[WARN ] : $msg_body"
                    : ( $msg_type eq 'INFO'  && $LOG_LEVEL >= 3 ) ? "[INFO ] : $msg_body"
                    : ( $msg_type eq 'DISPL' && $LOG_LEVEL >= 4 ) ? $msg_body
                    : ( $msg_type eq 'NOTE'  && $LOG_LEVEL >= 5 ) ? "[NOTE ] : $msg_body"
                    : ( $msg_type eq 'DEBUG' && $LOG_LEVEL >= 6 ) ? "\t[DEBUG] : $msg_body"
                    :                                               undef;

        if ( ! defined $msg_out && ! grep(/^$msg_type$/, ('ERROR', 'ALERT', 'WARN', 'INFO', 'DISPL','NOTE', 'DEBUG')) ) {
                $msg_out = "[SYS_ERROR] : Unrecognized message type [$msg_type]";
        }

	&logMsg("$msg_type","$msg_body") if ( $LOGINIT == 1 );

	if ( defined $msg_out ) {
		print $msg_out . "\n"; 
	}	

	return 0;
}

##################################################
# setLock : Locks running instance               #
##################################################
sub setLock() {

	&printMsg("INFO","Locking current execution instance");

	open($FDL,">>$LOCKFILE") or die "Failed to touch lockfile [$LOCKFILE] with error [$!]";
        unless ( flock($FDL, LOCK_EX | LOCK_NB) ) {
                &printMsg("ALERT","$0 is already running... Exiting ...");
                exit(1);
        }

	return 0;
}

##################################################
# setUnLock : Unlocks running instance           #
##################################################
sub setUnLock() {

	&printMsg("INFO","Unlocking current execution instance");

    if(openhandle($FDL)) {
        flock($FDL, LOCK_UN);
        close($FDL);
        unlink($LOCKFILE);
    }

	return 0;
}

####################################################
# setWorkingDir : set running directory to be local#
#                 used for non local path run      #
####################################################
sub setWorkingDir() {

	my $mypath = abs_path($0);
	my $mydir = dirname($mypath);

	&printMsg("DEBUG","Switching to [$mydir] working dir");

	eval {
		chdir $mydir or die "Failed to change working directory";
	};

	if ( $@ ) {
		&printMsg("ERROR","$@");
		return 1;
	}
	return 0;
}

#This subroutine works very similar to alarm except it is itself safe to use 
#with code that itself calls alarm or sleep. it does this by setting up a signal handler for SIGUSR1
#and then forking a process that waits the specified amount of time before giving us the SIGUSR1 signal. 
#Since forked processes do no inherit the alarm(2) timer it is safe to use this timer without resetting
#any previously set timers in the main process or vice-versa. We also restore anything previously saved into $SIG{USR1} if anything
#when the timer is reset with a 0 interval. Optionally if a subroutine reference is passed as the second argument
#it will be used as the SIGUSR1 handler instead of the default. 
{
my $child = 0;
my $oldusr1;
    sub setAlarm {
        my $interval = shift;
        my $handler = shift;
        my $defhandler = sub { die "ThreatStop::setAlarm() Timer Expired"; };
        #Set up signal handler
        if(!$handler) {
            $handler = $defhandler;
        }
        #Save existing USR1 handler if it exists.
        #but only do this once so we don't lose the original with multiple calls to setAlarm.
        if((! $oldusr1) and $SIG{USR1}) {
            if($SIG{USR1} != $defhandler) { #Make sure we don't "restore" our own internal handler later.
                &printMsg("DEBUG","ThreatStop::setAlarm() Saving existing USR1 handler for later");
                $oldusr1 = $SIG{USR1};
            }
        }                                                                                                                                                                                       
        $SIG{USR1} = $handler;                                                                                                                                                                  
                                                                                                                                                                                                
                                                                                                                                                                                                
        if(!$interval) { #handle unsetting the alarm and signal handler.
            &printMsg("DEBUG","ThreatStop::setAlarm() unsetting alarm");
            if($child) {
                &printMsg("DEBUG","ThreatStop::setAlarm() killing child $child");
                kill("TERM",$child);
                $child = 0;
            }
            if($oldusr1) {
                &printMsg("DEBUG","ThreatStop::setAlarm() restoring saved signal handler");
                $SIG{USR1} = $oldusr1;
                $oldusr1 = undef;
            } else { $SIG{USR1} = undef; }
            return 0;
        } else { #Fork the simple process to signal us at a later time.
            if($child) {
                &printMsg("DEBUG","ThreatStop::setAlarm() killing child $child");
                kill("TERM",$child);
            }
            my $pid = fork();
            if($pid) {
                &printMsg("DEBUG","ThreatStop::setAlarm() forked child $pid");
                $child = $pid;
                return 0;
            } else {
                &printMsg("DEBUG","ThreatStop::setAlarm() child setting alarm to $interval seconds");
                close($FDL) if openhandle($FDL); #alarm processes should not hold locks.
                sleep $interval;
                kill("USR1",getppid());
                exit(0);
            }
        }
}
}

##################################################
# Convert CIDR masks to dotted quads             #
##################################################
sub ipv4_cidr2msk($) {
        return join (".",  unpack 'CCCC', pack( "B*", ( "0" x shift ).( "1" x 32) ) );
}

##################################################
# os_detecter : Detects OS type                  #
##################################################
sub osPlatformDetecter() {

	my $res      = shift;
	my $res_full = shift;
	my $platform;

	$platform = qx(uname -m);
	$$res = 64 if ( $platform =~ /x86_64/ or $platform =~ /amd64/);
	$$res = 32 if ( $platform =~ /i686/   );

	chomp($platform);
	$$res_full = $platform;
	
	return 0;
}

############################################################
# osDestributionDetector : Detects OS version for Linux os #
############################################################
sub osDestributionDetector() {

    my $ver_ref = shift;

    eval {
        open (VFD, "/proc/version") or die "Cannot determine Linux Distribution\n";
    };
   
    if ( $@ ) {
        &printMsg("ERROR","$@");
        return 1;
    }

    $_ = join('',<VFD>);

    $$ver_ref = 'Unknown';
    $$ver_ref = 'RedHat' if (/Red Hat/m);
    $$ver_ref = 'Debian' if (/Debian/m);
    $$ver_ref = 'Ubuntu' if (/Ubuntu/m);

    &printMsg("DEBUG","Detected - [$$ver_ref]");

    return 0;
}

##################################################
# calculateElapsedTime :                         #
##################################################
sub calculateElapsedTime() {

	my $start_epoch      = shift;
	my $elapsed_time_ref = shift;
	my ($elapsed_time, $end_epoch);
        my ($hr, $min, $sec);

        $end_epoch = time();
        $elapsed_time = $end_epoch - $start_epoch;

        $sec = $elapsed_time % 60;
        $min = ( $elapsed_time / 60 ) % 60;
        $hr  = ( $elapsed_time / ( 60 * 60 )) % 24;

        $elapsed_time = sprintf("%02d:%02d:%02d",$hr,$min,$sec);
	$$elapsed_time_ref = $elapsed_time;
	
        return 0;
}

##################################################
# checkDNSaccess : Verifies DNS connection       #
##################################################
sub checkDNSaccess() {

	my $dig      = shift;
	my $dnslist  = shift;
	my $dnsok    = shift;
        my @dnslist  = split(/ /,$dnslist);
	my $dnsport  = shift;
        my $cmd;

        &printMsg("INFO","Checking DNS servers access ...");

	#If no dns list was received
	if ( ! @dnslist ) {
		&printMsg("ERROR","DNS servers list is not defined");
		return 1;
	}

	if ( ! -e $dig ) {
		&printMsg('ERROR',"Missing dig [$dig] tool file");
		return 1;
	}

	#Pass over dns servers list and try to connect
        foreach my $dns ( @dnslist ) {
	
                $cmd = "$dig -p $dnsport ptr \@$dns basic.threatstop.local 2>&1";
                &printMsg('INFO',"Verifying dns server [$dns] access");
                &printMsg('DEBUG',"$cmd");
                my $rc = qx($cmd);
		my $rv = $?;

		if ( $rv != 0 ) {

			#Parse dig output and extract error message
			chomp($rc);
			my @rc = split('\n',$rc);
			@rc = grep {/;; (?!global options)/} @rc;
			$rc = join(' ',@rc);
			$rc =~ s/;//g;
			chomp($rc);			

			&printMsg('WARN',"Failed in dig command with error [$rc]");
			next;
		}

                if ( $rc =~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s ) {
                        &printMsg('INFO',"$dns resolves OK");
                        push(@$dnsok, $dns);
                } else {
                        &printMsg("WARN","ThreatSTOP initial dig to $dns failed, skipping to the next server in the list.");
                }
        }

        if ( ! defined @$dnsok[0] ) {
                &printMsg("ERROR","All ThreatSTOP DNS servers unreachable");
                return 1;
        }

        &printMsg("INFO","Done checking DNS servers");
        return 0;
}

####################################################################################
# updateDNSlist :                                                                  #
#                 $DNS_DOMAIN_BASE = dns-servers.threatstop.local                  #
#                 $DNS_DOMAIN_ALT  = dns-servers.threatstop.com                    #
#                 $DNS_SERVERS_ALT = 8.8.8.8 8.8.4.4 208.67.220.220 208.67.222.222 #
#                 $DNS_PROTO       = +tcp                                          #
#                 $DNS_PROTO_ALT   = +notcp                                        #
#                 $DNS_MAX_TRIES   = 3                                             #
####################################################################################
sub updateDNSlist() {

	my $dig_tool        = shift;
        my $dns_list        = shift;
	my $dns_port        = shift;

        my $dns_dom         = $DNS_DOMAIN_BASE;
        my $dns_proto       = $DNS_PROTO;
        my $try_num         =  1;
        my $dns_local_lst;
        my $dns_results_lst;
        my $rc;

	#If not defined port param use default 53
	$dns_port = $DNS_PORT_BASE if ( ! defined $dns_port);

	#New addition. What happened if we are already running on nonstandard port
	#in this case update base port to be the port we start with
	$DNS_PORT_BASE = $dns_port if ( $dns_port != 53 );

	&printMsg("INFO","Updating DNS lists");

        #Try to get DNS servers list until or success is reached or all options failed  
        while ( 1 ) {
		
                #Do update DNS action. Make decesign based on return value 
                $rc = &updateDNSlistsDo($dig_tool,$dns_list, $dns_dom, $dns_proto, $dns_port, \$dns_results_lst);

                #Something wrong happened... abort ...
                if ( $rc == 1 ) {
			&printMsg("ERROR","DNS resolving operation failed ...");
                        return 1;
                }

		#All DNS servers in current list failed to respond
		#switch to alternate port
		if ( $rc == 2 ) {

			#If we are already running on alternate port no need to do this check
			if ( $DNS_PORT_BASE == 5353 ) {
				&printMsg("DEBUG","Already running on alternate port ... skipping to the next test...");
				$rc = 3;
			} else {
				$dns_port = $DNS_PORT_ALT;
				&printMsg("INFO","Switching to alternate port [$dns_port]");
                	        &printMsg("DEBUG","dns list  - [$dns_list]");
                        	&printMsg("DEBUG","dns_dom   - [$dns_dom]");
	                        &printMsg("DEBUG","dns_proto - [$dns_proto]");
				&printMsg("DEBUG","dns_port  - [$dns_port]");
                	        next;
			}
                }

                #All DNS servers in current list failed to respond
                #switch to alternate domain
                if ( $rc == 3 ) {
                        $dns_dom = $DNS_DOMAIN_ALT;
			$dns_port = $DNS_PORT_BASE;
                        &printMsg("INFO","Switching to alternate domain [$dns_dom]");
                        &printMsg("DEBUG","dns list  - [$dns_list]");
                        &printMsg("DEBUG","dns_dom   - [$dns_dom]");
                        &printMsg("DEBUG","dns_proto - [$dns_proto]");
			&printMsg("DEBUG","dns_port  - [$dns_port]");
                        next;
                }

                #All DNS servers in current list failied to respond and
                #also alternate domain had been tried. Switch to alt
                #servers list
                if ( $rc == 4 ) {

			#From this moment we are going to global testing environment. This means
			#that if we have used as base port some alternate port we need to swich our
			#requests now to standard defaukt 53 port
			if ( $DNS_PORT_BASE != 53 ) {
				&printMsg("DEBUG","Running on non standard dns port [$DNS_PORT_BASE] switching to [53] port");
				$DNS_PORT_BASE = 53;
			}

                        $dns_list = $DNS_SERVERS_ALT;
                        $dns_dom  = $DNS_DOMAIN_ALT;
                        $dns_port = $DNS_PORT_BASE;

                        &printMsg("INFO","Switching to alternate global servers list [$dns_list]");
                        &printMsg("DEBUG","dns list  - [$dns_list]");
                        &printMsg("DEBUG","dns_dom   - [$dns_dom]");
                        &printMsg("DEBUG","dns_proto - [$dns_proto]");
			&printMsg("DEBUG","dns_port  - [$dns_port]");
                        next;
                }

                #OK, all servers from the GLOBAL servers list had been tried 
                #and failed trying to fallback to nameservers list defined on 
                #the box resolve.conf file
                if ( $rc == 5 ) {

                        &readLocalBoxNameserversLst(\$dns_local_lst);
                        $dns_dom         = $DNS_DOMAIN_ALT;
                        $dns_list        = $dns_local_lst;
                        $DNS_SERVERS_LCL = $dns_local_lst;
			$dns_port        = $DNS_PORT_BASE;

                        &printMsg("INFO","Switching localy defined nameservers list");
                        &printMsg("DEBUG","dns list  - [$dns_list]");
                        &printMsg("DEBUG","dns_dom   - [$dns_dom]");
                        &printMsg("DEBUG","dns_proto - [$dns_proto]");
			&printMsg("DEBUG","dns_port  - [$dns_port]");
                        next;
                }

                #All checks failed. Last thing to do is to fall back to UDP protocol and
                #re-do whole process again.  
                if ( $rc == 6 ) {
                        $dns_list  = $dns_server;
                        $dns_dom   = $DNS_DOMAIN_BASE;
                        $dns_proto = $DNS_PROTO_ALT;
                        $dns_port  = $DNS_PORT_BASE;
                        &printMsg("INFO","Switchibg to UDP protocol");
                        &printMsg("DEBUG","dns list  - [$dns_list]");
                        &printMsg("DEBUG","dns_dom   - [$dns_dom]");
                        &printMsg("DEBUG","dns_proto - [$dns_proto]");
			&printMsg("DEBUG","dns_port  - [$dns_port]");
                        next;
                }

                #We have tried all means and failed ... Abort futher execution ...
                if ( $rc == 7 ) {
                        &printMsg("ERROR","Failed to retieve updated DNS servers list");
                        return 1;
                }

                #The servers respond ok but returned DNS list is different
                #than original one. Need to re-run previous procedure with
                #new list settings
		if ( $rc == 9 ) {
                        #Max number of tries reached... Abort futher execution ...
                        if ( $try_num > $DNS_MAX_TRIES ) {
                                &printMsg("ERROR","Maximun number of tries to resolve DNS servers list had been reached...aborting");
                                return 1;
                        }

                        #Run again with new list
                        &printMsg("DEBUG","Re-running query process again with new DNS list... [${try_num}/${DNS_MAX_TRIES}]");
                        $dns_list   = $dns_results_lst;
                        $try_num++;
                        next;
                }

                #Everyting ok and lists are similar break the loop
                if ( $rc == 8 ) {
                        &printMsg("DEBUG","DNS list check completed");
                        last;
                }
        }

        #Now if DNS list had been changed and we successfully passed verification process update
        #the threatstop.conf file and $dns_server setting with new DNS data
        if ( $dns_results_lst ne $dns_server) {
                $dns_server = $dns_results_lst;
        }

	#If we are running on alternate port
	#if ( $dns_port != $DNS_PORT_BASE ) {
	#	$DNS_PORT = $dns_port;
	#}

        &printMsg("INFO","Updating config file with new DNS servers list");
        return 1 if ( &updateDNSConfFile() != 0 );
        return 0;
}

##########################################################
# loadConfigFileData : Load previous setup configuration #
##########################################################
sub loadConfFileData() {

	my $confile   = shift;
	my $data_hash = shift;
	my $err_msg   = shift;
	my $FILE;
    my $basedir = $CONFIG_PATH;

        #Reading configuration file
        &printMsg("INFO","Initializing data from configuration file");

        if( -e $confile ) {

            eval {
                    open (FILE, "<", $confile) or die "$!";
            };
        } else {
            eval {
                    open (FILE, "<", "$basedir/$confile") or die "$!";
                    $confile = "$basedir/$confile";
            };
        }

        if ( $@ ) {
                $$err_msg = "$@";
                return 1;
        }

        #Parse data
        while (<FILE>) {
                next unless (/^([a-z]\w+)=[("]?(.*?)[")]?\s+$/);
                $data_hash->{$1}=$2;
        }
        close FILE;
        mpw($confile,$data_hash);
	#Also update DNS servers list with pre-stored setting
	$DNS_LIST = $dns_server if ( defined $dns_server );

        return 0;
}

###############################################################
##################### MODULE END ##############################
###############################################################

1;
