#!/usr/bin/perl 

# /**
#  *
#  * Copyright(c) 2006-2016 ThreatSTOP, Inc. All Rights Reserved
#  *
#  * NOTICE: All information contained herein is, and remains the property
#  * of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and 
#  * technical concepts contained herein are proprietary to ThreatSTOP, Inc. 
#  * and its suppliers and certain aspects thereof are protected by United 
#  * States Patent No. 8,533,822 and United States Patent No. 8,869,237, 
#  * and are protected by trade secret or copyright law.
#  *
#  * US Government Entities: The ThreatSTOP service, software and
#  * documentation, as applicable, are "commercial computer software" and
#  * "commercial computer software documentation" developed exclusively at
#  * private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
#  * 12.212 or DFARS 227 7202 and their successors, as applicable, use,
#  * reproduction and disclosure of the software and documentation is
#  * governed by the terms of Threatstop's commercial agreement.
#  */

use strict;
use warnings;

use LWP::UserAgent;
use HTTP::Request::Common;
use Digest::MD5;
use Cwd;
use Cwd qw(abs_path);
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/lib/";
use ThreatStop;

############################################################
################### GLOBAL DEFENITIONS PART ################
############################################################

my $MYDIR = "$FindBin::Bin";

my $CTIME;
my $ETIME;
my $START_EPOCH   = time();
my $CFG           = {};								# Holds configuration settings
my $MAX_TRIES     = 5;								# If could not retrive data try again
my $MAX_LOG_SIZE  = 10000;							# Cut logs in 10000 chunks
my $CHUNK_COUNTER = 0;								# Holds number of chunk we devide logfile
my $LOGINIT       = 1;								# Enable operation logging
my $LOGFILE       = $MYDIR . "/ts-logupload.log";	# Store operation log here
my $RESLOG        = "/tmp/ts-logupload.txt";    	# Store execution log here 
my $TMPLOGFILE    = "/tmp/tmplog";					# Tmp log file holding 10000 lines chunk
my $CONFIG_PATH   = "/etc/threatstop";
chomp (my $VERSION = `cat $CONFIG_PATH/VERSION`);
my $CONFILE       = $CONFIG_PATH . "/threatstop.conf";	# Configuration settings file
my $IP;

$DEBUG_LEVEL      = 6;								# Defines verbose level of operation

############################################################
################### FUNCTIONS PART #########################
############################################################

###################################################
# verifyCMDdata : verify supplied CMD data format # 
#                 ip - format parameter check     #
###################################################
sub verifyCMDdata() {

	&printMsg("INFO","Verifying supplied CMD data format");

	return 0 if (! defined $ARGV[0] );
	return 0 if ($ARGV[0] eq "DEBUG" );

	#See is our arg[0] has ip format
	$IP = $ARGV[0];
    unless ( $IP =~ /^\d+\.\d+\.\d+\.\d+$/ ) {
    	&printMsg("ERROR","$IP not a valid ipaddress");
        return 1;
    }

	return 0;
}

#################################################################
# loadConfigData : load configuration settings from config file #
#################################################################
sub loadConfigData() {

	my $err_msg;

	&printMsg("INFO","Loading previous configuration");

	if ( &loadConfFileData($CONFILE,$CFG, \$err_msg) != 0 ) {
		&printMsg("ERROR","Failed to load data from config file with error [$err_msg]");
		return 1;
	}

	return 0;
}


########################################################
# cleanUpPrevData : Remove possible leftover tmp files #
########################################################
sub cleanUpPrevData() {

	my $cmd = "ls /tmp/tmplog_*.log 2>/dev/null | wc -l";
	my $res = qx($cmd);
	my $cntr;

	for($cntr=0; $cntr<$res; $cntr++) {
		&printMsg("DEBUG","Cleaning [/tmp/tmplog_${cntr}.log] previous tmp file");
		unlink("/tmp/tmplog_${cntr}.log");
	}

	return 0;
}

########################################################################
# verifyLogFile : Verify log file stats, if empty look for rotated log #
########################################################################
sub verifyLogFile() {

	my $logfile = $CFG->{'logfile'};

	&printMsg("INFO","Verifying log file [$logfile] stats");

	#If logfile does not exists abort
	if ( ! -e "$logfile" ) {
                &printMsg("WARN","Log file [$logfile] does not exists... no data will be send...");
		return 1;
        }

	#If logfile is empty see if we have it's previous rotated instance
	my $log_size = (stat("$logfile"))[7];
        if ( $log_size == 0 ) {
		
		&printMsg("INFO","Log file [$logfile] has zerro length... Switching for previous rotated instance");
		$logfile = $logfile . ".1";

		#If no rotated instance abort
		if ( ! -e "$logfile" ) {
			&printMsg("ERROR","No rotated instance of log file [$logfile] had been found");
			return 1;
		}	

		#If rotated instance is empty abort
		$log_size = (stat("$logfile"))[7];
		if ( $log_size == 0 ) {
			&printMsg("INFO","Log file rotated instance [$logfile] has zerro length... Aboritng ...");
			return 1;
		}

		#Set logfile to be rotated instance which actually will be processed
		&printMsg("DEBUG","Pointing logfile name to rotated instance $logfile -> $CFG->{'logfile'}");
		$CFG->{'logfile'} = $logfile;
	}

	return 0;
}

##############################################################################
# processLogFile : process log file. Separate it in MAX_LOG_SIZE chunks than #
#                  upload each chunk separetly                               #
#                  MAX_LOG_SIZE = 10000 lines 								 #
##############################################################################
sub processLogFile() {

	my $line_ctr = 0;

	&printMsg("INFO","Processing [$CFG->{'logfile'}] log file into tmp chunks with predefined [$MAX_LOG_SIZE] size");

	#Open logfile for reading. We will want to separate it in 100000 lines chunks
	eval {
		open(LFH,"$CFG->{'logfile'}") or die "Failed to open log file [$CFG->{'logfile'}] with error [$!]";
	};
	
	#Take care of possible errors
	if ( $@ ) {
		&printMsg("ERROR","$@");
		return 1;
	}

	#Pass over log data and separate it into chunks	
	while ( <LFH> ) {
		$line_ctr++;

		if ( $line_ctr == $MAX_LOG_SIZE ) {
			#Max chunk size reached next line will be written itno newtmp file
			$line_ctr = 0;

			#Write line itno current tmp file
			print TFH;		

			#Close current tmp file. We will start new tmp file on next itteration
			close TFH;

			#Go for next itteration
			next;

		} 

		if ( $line_ctr == 1 ) {

			&printMsg("DEBUG","Writng into [${TMPLOGFILE}_${CHUNK_COUNTER}.log] tmp file");

			#Start new tmp file, first chunk line
			eval {
				open(TFH,">${TMPLOGFILE}_${CHUNK_COUNTER}.log") or die "Failed to open tmp log file [${TMPLOGFILE}_${CHUNK_COUNTER}.log] with error [$!]";
			};

			#Abort if we failed to open tmp file
			if ( $@ ) {
				&printMsg("ERROR","$@");
				return 1;
			}

			#Increase chunk counter 
			$CHUNK_COUNTER++;

		}

		#In any case write current line into current tmpfile 
		print TFH;		
	
	}

	#Close LOG file we finish reading it into tmp files
	close LFH;
	close TFH;

	return 0;
}

###################################################
# prepConnectionParams : prep environment settings#
#                        used in connection       #
###################################################
sub prepConnectionParams() {

	my $tmplogfil    = shift;
	my $filesize_ref = shift;
	my $local_md5    = shift;

	# build MD5 as hexdigest
    my $ctx = Digest::MD5->new;

	&printMsg("INFO","Preparing connection data");

	#Prepare to write data into file 
	eval {
		open( MD, $tmplogfil ) or die "Failed to open tmp file [$tmplogfil] with error [$!]";
	};

	if ( $@ ) {
		&printMsg("ERROR","$@");
	}

    $ctx->addfile(*MD);
    $$local_md5 = $ctx->hexdigest;

    close MD;

	# calculate file size
	$$filesize_ref = ( stat $tmplogfil )[7];

	return 0;
}

#########################################################
# sendDataDo : Try to send data from supplied tmp file  #
#########################################################
sub sendDataDo() {

	my $tmplogfil = shift;
	my $try       = shift;
	my $filesize;
	my $local_md5;
	my $req;

	#Prepare connection test parameters
	&prepConnectionParams($tmplogfil,\$filesize,\$local_md5);

	&printMsg("DEBUG","Sending data ....");

	my $ua = LWP::UserAgent->new();

	# put timeouts, proxy etc into the useragent if needed
    if ($IP) {

		&printMsg("DEBUG","fname - $tmplogfil"); 
		&printMsg("DEBUG","fsize - $filesize"); 
		&printMsg("DEBUG","md5   - $local_md5");
		&printMsg("DEBUG","ip    - $IP");

		$req = 	POST $CFG->{'url'},
			   	Content_Type => 'form-data',
				Content      => [
					submit      => 1,
					upfile_size => $filesize,
					upfile      => [$tmplogfil],
					md5_client  => $local_md5,
					fw_ip       => $IP
				];
	} else {

		&printMsg("DEBUG","fname - $tmplogfil"); 
		&printMsg("DEBUG","fsize - $filesize ");
		&printMsg("DEBUG","md5   - $local_md5");
		&printMsg("DEBUG","url   - $CFG->{'url'}");

		$req = 	POST $CFG->{'url'},
				Content_Type => 'form-data',
				Content      => [
					submit      => 1,
					upfile_size => $filesize,
					upfile      => [$tmplogfil],
					md5_client  => $local_md5
				];
	}

	my $response = $ua->request($req);

	#Now analyse responseonse and return relevant return code
	if ( $response->is_success() ) {

		if ( $response->content =~ /MD5: OK\s+Filesize: OK/s ) {
			&printMsg("INFO","Upload was succesfull [" . $response->status_line . "]");	
			return 0;
		}

		if ( $response->content =~ /Upload Error/s ) {
			&printMsg("ERROR","Upload failed with response \n[" . $response->as_string . "]");
			return 1;
		}

    	&printMsg("WARN","Connection succesfull but could not get the proper return state [" . $response->status_line . "]"); 
    	#&printMsg("WARN","Connection succesfull but could not get the proper return state [" . $response->as_string . "]"); 

		open(OUT,">$RESLOG");
		print OUT "Probably OK on $try try", $response->content . $/;
		close OUT;
		return 2;
	}

	#If we are here the connection failed
	&printMsg("ERROR","Conection failed with response [" . $response->status_line . "]");
	#&printMsg("ERROR","Conection failed with response [" . $response->as_string . "]");
	return 1;
}

##############################################################
# sendTmpFileData : try to send data from supplied tmp file. #
#                   return failure if status was not ok      #
##############################################################
sub sendTmpFileData() {

	my $cntr     = shift;
	my $try      = shift;
	my $rv       = 0;
	my $tmp_file = ${TMPLOGFILE} . "_" . ${cntr} . ".log";

	#Try to send data
	$rv = &sendDataDo($tmp_file, $try);

	return $rv;
}

#####################################################################################
# sendProcessedData : send data stored in tmp files we created in previous function #
#####################################################################################
sub sendProcessedData() {

	my $cntr;
	my $tries = 0;
	my $rv;

	&printMsg("INFO","Start sending data");

	#Pass over each tmp file we have created and send it
	for($cntr=0; $cntr<$CHUNK_COUNTER; $cntr++ ) {
		&printMsg("DEBUG","Sending ${TMPLOGFILE}_${cntr}.log file");
		$rv = &sendTmpFileData($cntr,$tries);

		if ( $rv == 0 ) {
			#Send data was ok, go to the next tmp file
			next;
		}

		if ( $rv == 1 ) {
			#Connection failed, abbort
			&printMsg("ERROR","Connection failed....Aborting....");
			return 1;
		}

		
		if ( $rv == 2 ) {
			#Connection established but send status was not ok
			if ( $tries == $MAX_TRIES ) {
				#Max number of tires reached. Aborting with error
				&printMsg("ALERT","Max number of connection tries was reached... Aborting...");
				return 1;
			}

			#Try to re-send current tmp file again
			&printMsg("WARN","Trying to re-send ${TMPLOGFILE}_${cntr}.log on attempt [${tries}\/$MAX_TRIES]");
			$tries++;
			$cntr--;
		}
	}

	return 0;
}

############################################################
#################### MAIN PART #############################
############################################################

&getCurrentTime(\$CTIME);
&logInit($LOGFILE);
&printMsg("INFO","Starting ThreatSTOP logupload operation v$VERSION at $CTIME");

exit 1 if ( &cleanUpPrevData()   != 0 );    # Remove previous run tmp files
exit 1 if ( &verifyCMDdata()     != 0 );	# If CMD params were supplied verify format
exit 1 if ( &loadConfigData()    != 0 );	# Load configuration from threatstop.conf file 
exit 1 if ( &verifyLogFile       != 0 );    # Verify log file paramters and content 
exit 1 if ( &processLogFile()    != 0 );	# Read log data into tmp files of MAX_LOG_SIZE lines
exit 1 if ( &sendProcessedData() != 0 );	# Send data stored in tmp files
exit 1 if ( &cleanUpPrevData()   != 0 );    # Remove previous run tmp files

&getCurrentTime(\$CTIME);
&calculateElapsedTime($START_EPOCH,\$ETIME);
&printMsg("INFO","Finish ThreatSTOP logupload operation at $CTIME after $ETIME");

