#!/bin/bash

# /**
#  *
#  * Copyright(c) 2006-2016 ThreatSTOP, Inc. All Rights Reserved
#  *
#  * NOTICE: All information contained herein is, and remains the property
#  * of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and 
#  * technical concepts contained herein are proprietary to ThreatSTOP, Inc. 
#  * and its suppliers and certain aspects thereof are protected by United 
#  * States Patent No. 8,533,822 and United States Patent No. 8,869,237, 
#  * and are protected by trade secret or copyright law.
#  *
#  * US Government Entities: The ThreatSTOP service, software and
#  * documentation, as applicable, are "commercial computer software" and
#  * "commercial computer software documentation" developed exclusively at
#  * private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
#  * 12.212 or DFARS 227 7202 and their successors, as applicable, use,
#  * reproduction and disclosure of the software and documentation is
#  * governed by the terms of Threatstop's commercial agreement.
#  */

###############################################################
################ GLOBAL DEFINITIONS PART ######################
###############################################################


PATH=$PATH:/sbin

SCRIPTS_PATH=/opt/threatstop
CONFIG_PATH=/etc/threatstop
START_EPOCH=`date -R`
CRONTAB=$(which crontab)

VERSION=$( cat $CONFIG_PATH/VERSION )

LOCKFILE="setup.lck"
LOGFILE="tsoutput.log"
CONFILE="${CONFIG_PATH}/threatstop.conf"
UPLOADFILE="ts-a10.pl"
DIG_TOOL=$(which dig 2>/dev/null)
MAXPOLICYSIZE=10000
MAXPOLICYALLOW=65535
BAR_LENGTH=35
PORT=53
BASE='/etc/threatstop/lists'

TMP_CRONFILE="${CONFIG_PATH}/tmp_cron.cron"

TEMPLATES_PATH="${CONFIG_PATH}/files"
TEMPLATE_CONFILE="threatstop.conf.example"
TEMPLATE_SYSLOG_ROTATE_FILE="threatstop-logrotate.conf"
TEMPLATE_SYSLOGNG_CONF_FILE="01-threatstop-syslog-ng.conf"
TEMPLATE_SYSLOGR_CONF_FILE="threatstop-rsyslog.conf"
TEMPLATE_SYSKLOG_CONF_FILE="threatstop-syslog.conf"
SYSTEM_SYSKLOG_CONF_FILE="syslog.conf"

SYSLOG_CRON_PATH="/etc/cron.d"
SYSLOG_CRON_ROTATE_FILE="root_cron"
SYSLOG_CRON_UPDATE_FILE="threatstop_cron"
SYSLOG_ROTATE_SYSTEM_FILE="/etc/logrotate.d/remotes"
SYSLOGNG_CONF_SYSTEM_PATH="/etc/syslog-ng/conf.d"
SYSLOGR_CONF_SYSTEM_PATH="/etc/rsyslog.d"
SYSKLOG_CONF_SYSTEM_PATH="/etc"
TARGET_LOG=/var/log/remotes.log
FIREWALLD_RSYSLOG_SERVICE_FILE=/etc/firewalld/services/rsyslog.xml

CLI_IP=""
CLI_BLOCKLST=""
CLI_ALLOWLST=""
CLI_PORT=""
CLI_MAXPOLICYSIZE=""

MANDATORY_PARAMS_LIST=(dns_server internal_ip external_ip block_list allow_list logfile url maxpolicysize port)

MANDATORY_FILES_LIST=(${SCRIPTS_PATH}/$UPLOADFILE ${TEMPLATES_PATH}/${TEMPLATES_CONFILE} ${TEMPLATES_PATH}/${TEMPLATE_SYSLOG_ROTATE_FILE} ${TEMPLATES_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} ${TEMPLATES_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE})

PERL_MOD_LIST=('Net::CIDR::Lite' 'HTTP::Request::Common' 'Net::SSLeay' 'Fcntl' 'LWP')
DEBUG="false"
[ "$1" = "DEBUG" ] && DEBUG="true" && shift

#Changing to working dir
DIR=$(dirname $0)
[ "$DIR" != '.' ] && cd $DIR

###############################################################
#################### FUNCTIONS PART ###########################
###############################################################

############################################
function aborting() {

        unLock

        echo 
        echo "****************************************************************"
        echo "*Unrecoverable errors encountered... aborting...               *" 
        echo "*Please fix, then run script again, or contact support for     *"
        echo "*assistance : support@threatstop.com                           *"
        echo "****************************************************************"
        echo
        exit 1
}

############################################
function echoMsg() {

        case $1 in
                ERROR) echo "[ERROR] : $2"
                        ;;
                WARN)  echo "[WARN ] : $2"
                        ;;
                ALERT) echo "[ALERT] : $2"
                        ;;
                INFO)  echo "[INFO ] : $2"
                        ;;
                DEBUG) if [ $DEBUG == true ]; then
                                echo -e "\t[DEBUG] : $2"
                       fi
                        ;;
                *) echo "SCRIPT ERROR : Unrecognized parameter [$1]"
        esac
}

############################################
function setLock() { 
          
        echoMsg INFO "Locking current setup session flow"
          
        if [ -f "$LOCKFILE" ]; then
                PID=$(cat $LOCKFILE)
                echoMsg WARN "Operation is locked by another process with pid [$PID].. exiting"
                return 1
        fi   

        #Make sure we clean up in case of user iitiates interrupts
        trap "{ rm -f $LOCKFILE; rm -f $TMP_CRONFILE; echo; exit 255; }" EXIT SIGINT SIGTERM
    
        echo $$ > $LOCKFILE
        return 0
}  

############################################
function unLock() {

        echoMsg INFO "Releasing the lock"
        rm -f $LOCKFILE
}

############################################
function isIPok() {

        local ip=$1
        local state=1

        if [[ $ip =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]; then
                ips[0]=$(echo $ip | cut -d\. -f1)
                ips[1]=$(echo $ip | cut -d\. -f2)
                ips[2]=$(echo $ip | cut -d\. -f3)
                ips[3]=$(echo $ip | cut -d\. -f4)

                [[ ${ips[0]} -le 255 && ${ips[1]} -le 255 && ${ips[2]} -le 255 && ${ips[3]} -le 255 ]] && state=0
        fi

	[ $state -eq 1 ] && echoMsg ERROR "Bad ip [$ip] format" && return 1
        return 0
}

############################################
function isPolicyOk() {

        local pol=$1

        #Check to be a number
        if ! [[ $pol =~ ^[0-9]+$ ]]; then
                echoMsg ERROR "Wrong supplied MAX policy size format [$pol]"
                return 1
        fi

        #check for size exceeding
        if [ $pol -gt $MAXPOLICYALLOW ]; then
                echoMsg ERROR "Bad MAX policy size must be less than $MAXPOLICYALLOW"
                return 1
        fi

        return 0
}

############################################
function showPrepMsg() {

echo
echo "*******************************************************************"

cat <<MSG
Welcome to the ThreatSTOP setup for A10 Controllers v$VERSION

At the first stage of this setup process you will be given the chance 
to specify a number of setup options. You can always change them 
later by re-running the setup script or by manually updating the 
relevant entries in threatstop.conf file.

If a parameter had been previously configured, it's value will be shown
in [] on parameter request line. You can choose to re-use this value 
by simply pressing enter.
MSG

echo "*******************************************************************"
echo

}

############################################
function rollBackSysConfig() {

        echoMsg ALERT "Rolling back system configuration"

        if [ -f "${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE}" ]; then
                echoMsg DEBUG "Removing ${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} file"
                sudo rm -f ${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE}
        fi

        if [ -f "${SYSLOGR_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE}" ]; then
                echoMsg DEBUG "Updating ${SYSLOGR_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE} file"
                sudo rm -f ${SYSLOGR_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE}
        fi

        if [ -f "${SYSLOG_ROTATE_SYSTEM_FILE}" ]; then
                echoMsg DEBUG "Updating ${SYSLOG_ROTATE_SYSTEM_FILE} file"
                sudo rm -f ${SYSLOG_ROTATE_SYSTEM_FILE}
        fi

        echoMsg DEBUG "Cleaning up update crontab entry"
        $CRONTAB -l 2>/dev/null > $TMP_CRONFILE
        sed -i "/$UPLOADFILE/d" $TMP_CRONFILE
        rc=$($CRONTAB $TMP_CRONFILE 2>&1)
        if [ $? -ne 0 ]; then
                echoMsg ERROR "Failed to remove upload cron job entry"
        fi

        echoMsg DEBUG "Cleaning up rotate crontab entry"
        sudo $CRONTAB -l 2>/dev/null > $TMP_CRONFILE

        sed -i '/exec q(\/usr\/sbin\/logrotate/d' $TMP_CRONFILE
        sed -i '/Force a logrotate if the log/d' $TMP_CRONFILE
        sed -i "/$UPLOADFILE/d" $TMP_CRONFILE
        sed -i "/^$/d" $TMP_CRONFILE
        rc=$(sudo $CRONTAB $TMP_CRONFILE 2>&1)
        if [ $? -ne 0 ]; then
                echoMsg ERROR "Failed to remove reload cron job entry"
        fi

        echoMsg INFO "System state has been rolled back"

        return 0
}

############################################
function usage() {

        echo
        echo -e "\t*****************************************************************************************************\n"
        echo -e "\tUsage: $0 [-i] [-s] [-b|--blocklist blocklist.threatstop.local] ] [-a|--allowlist allowlist.threatstop.local] ] [-h|--help] [-m|maxpolicysize] [-p|port]"
        echo -e "\t*****************************************************************************************************\n"

        unLock
        echo
        exit 1
}


############################################
function verifySystemCfg() {

        echoMsg INFO "Verifying required scripts and templates existance"
        echo
        local dots=""
        local isOK=1

        for item in ${MANDATORY_FILES_LIST[*]}; do
                if [ -e "$item" ]; then

                        #For each item create a dot line of the same size
                        local len=${#item}
                        len=$(($BAR_LENGTH - $len))

                        for((i=0; i<$len; i++)); do
                                dots="${dots}."
                        done

                        echo -e "\tVerifying file [$item] $dots [OK]"
                else
                        echo -e "\tVerifying file [$item] $dots [FAIL]"
                        isOK=0
                fi
                dots=""
        done

        if [ $isOK -eq 0 ]; then
                echoMsg ERROR "Missing mandatory files/templates.... "
                return 1
        fi

	# Verify dig utility existance
	if [ -z "$DIG_TOOL" ] || [ ! -e "$DIG_TOOL" ]; then
		echoMsg ERROR "Could not detect dig tool existance"
		return 1
	fi

        echo
        echoMsg INFO "Mandatory files verification passed"
        echoMsg INFO "Verifying required Perl module existance"
        echo

	for mod in ${PERL_MOD_LIST[*]}; do

                #For each module create the dor line of the same size
                local len=${#mod}
                len=$(($BAR_LENGTH - $len - 5))

                for((i=0; i<$len; i++)); do
                        dots="${dots}."
                done

                echo -ne "\tVerifying perl module $mod $dots ";
                perl -I"${SCRIPTS_PATH}/lib" -M${mod} -e 1 &>/dev/null
                if [ $? -ne 0 ]; then
                        echo "[FAIL]"
                        isOK=0
                else
                        echo "[OK]"
                fi
                dots=""
        done

        echo
        if [ $isOK -eq 0 ]; then
                echoMsg ERROR "Missing mandatory Perl modules. Please install them first"
                return 1
        fi

        echoMsg INFO "Perl modules verification passed"
        return 0
}

#############################################################
function loadPrevSettings() {

	echoMsg INFO "Loading default configuration from template"
        . ${CONFIG_PATH}/$TEMPLATE_CONFILE

	if [ -e "$CONFILE" ]; then
		echoMsg INFO "Loading previous configuration"
		. $CONFILE
	fi

	#Add default settings if  missing in confile
	internal_ip=$device
	[ -z "$maxpolicysize" ] && maxpolicysize=$MAXPOLICYSIZE
	[ -z "$port" ] && port=$PORT
	[ -z "$base" ] && base=$BASE

	return 0
}

#############################################################
function readCLIParams() {

	if [ "$ddos" -eq 1 ]; then
		echoMsg INFO "CLI not supported in DDOS mode"
		return 0;
	fi

	echoMsg INFO "Verifying user supplied parameters"

	while [ ! -z $1 ]; do

                case $1 in
                        -b|-blocklist)
                                local list=$2
                                if ! echo "$list" | grep 'threatstop.local' &>/dev/null; then
                                        #Add threatstop.local to list name
                                        list="${list}.threatstop.local"
                                fi
                                CLI_BLOCKLST=$list
                                shift
                                ;;
                        -a|-allowlist)
                                local list=$2
                                if ! echo "$list" | grep 'threatstop.local' &>/dev/null; then
                                        #Add threatstop.local to list name
                                        list="${list}.threatstop.local"
                                fi
                                CLI_ALLOWLST=$list
                                shift
                                ;;
                        -i|-ip)
                                local ip=$2
                                isIPok "$ip" 
                                [ $? -eq 0 ] && return 1
                                CLI_IP=$ip
                                shift
                                ;;
                        -m|-maxpolicysize)
                                local policysize=$2
                                isPolicyOk $policysize
                                [ $? -eq 0 ] && return 1
                                CLI_MAXPOLICYSIZE=$policysize
                                shift
                                ;;
			-p|port)
				CLI_PORT=$2
				shift
				;;
                        -h|--help|-?)
				usage
                                ;;
                        *)echoMsg ERROR "Unrecoginzed parameter [$1] supplied"
                          usage
                                ;;
                esac
                shift
        done

	return 0
}

##############################################################
function userAsk() {

        #If user had supplied CLI parameters use them instead of asking

        echo

        if [ "$ddos" -eq 1 ]; then

                if [ -z "${dns_server_rpz}" ]; then
                        echo "Please enter DNS server to get lists from : "
                        read -p "[${dns_server_rpz}] ==> " opt
                        [ ! -z "$opt" ] && dns_server_rpz=$opt && unset opt
                        echo
                fi

                MANDATORY_PARAMS_LIST=(dns_server ddos ddos_zone_candidates ddos_zonelist dns_server_rpz internal_ip external_ip block_list allow_list logfile url maxpolicysize port)

                # loop over zone selections
                ddos_zonelist="" # start clean
                echo "DDOS zones :"
                for zoneinfo in "${ddos_zone_candidates[@]}"; do

                        zoneprompt=${zoneinfo%%:*}   # what's before the colon
                        zonetype=${zoneinfo##*:}     # what's after the colon
                        read -p "Enable $zoneprompt (y/n) ? [y] ==> " opt
                        [ -z "$opt" ] && opt=y;
                        optchar=$(echo $opt | cut -b 1)
                        if [ $optchar = "y" -o $optchar = "Y" ]; then
                                if [ -z "$ddos_zonelist" ]; then
                                        ddos_zonelist=${zonetype}
                                else
                                        ddos_zonelist="${ddos_zonelist} ${zonetype}"
                                fi
                        fi
                done
                echo
        fi

        if [ -z "$dns_server" ]; then
                echo "Please enter DNS servers list to get lists from :"
                read -p "[$dns_server] ==> " opt
                [ ! -z "$opt" ] && dns_server=$opt && unset opt
                echo
        fi

        #Use supplied CLI block or ask
        if [ -z "$CLI_BLOCKLST" ]; then
                echo "Please enter the block_list to use: "
                read -p "[$block_list] ==> " opt
                [ ! -z "$opt" ] && block_list=$opt && unset opt
                echo
        else
                block_list=$CLI_BLOCKLST
        fi

        #Use supplied CLI allow_list or ask
        if [ -z "$CLI_ALLOWLST" ]; then
                echo "Please enter the allow_list to use: "
                read -p "[$allow_list] ==> " opt
                [ ! -z "$opt" ] && allow_list=$opt && unset opt
                echo
        else
                allow_list=$CLI_ALLOWLST
        fi

        #Use supplied CLI ip or ask
        if [ -z "$CLI_IP" ]; then
                echo "Please enter the external IP address of the A10 device: "
                read -p "[$external_ip] ==> " opt
                [ ! -z "$opt" ] && external_ip=$opt && unset opt
                echo
        else
                external_ip=$CLI_IP
        fi

        #Use supplied CLI ip or ask
        if [ -z "$CLI_PORT" ]; then
                echo "Please enter port to use for DNS queries : "
                read -p "[$port] ==> " opt
                [ ! -z "$opt" ] && port=$opt && unset opt
                echo
        else
                port=$CLI_PORT
        fi

	echo "Please enter the internal IP address of the A10 device: "
        read -p "[$internal_ip] ==> " opt
        [ ! -z "$opt" ] && internal_ip=$opt && unset opt
        echo

        echo "Please enter the directory to store the A10 class lists: "
        read -p "[$base] ==> " opt
        [ ! -z "$opt" ] && base=$opt && unset opt
        echo

        #Ask only if we are missing this value
        if [ -z "$logfile" ]; then
                echo "Please enter log upload script placement: "
                read -p "[$logfile] ==> " opt
                [ ! -z "$opt" ] && logfile=$opt && unset opt
                echo
        fi

	#Ask only if we are missing this value
        if [ -z "$url" ]; then
                echo "Please enter URL to upload logfiles to : "
                read -p "[$url] ==> " opt
                [ ! -z "$opt" ] && url=$opt && unset opt
                echo
        fi

        echo

	return 0
}


##############################################################
function readUserData() {

	echoMsg INFO "Entering interactive stage"

        while [ 1 ]; do
                userAsk

		local sp="*******"
                local ep="*******"
                [ -z "$password" ] && sp="NOT SET"
                [ -z "$enable_pw" ] && ep="NOT SET"

                echo "Chosen configuration parameters"
                echo "--------------------------------"
                echo "A10 device external IP : $external_ip"
                echo "A10 device internal IP : $internal_ip"
                # echo "Username                 : $username"
                # echo "SSH password             : $sp"
                echo "ThreatSTOP block list    : $block_list"
                echo "ThreatSTOP allow list    : $allow_list"
		echo "Port                     : $port"
		echo "Class list directory     : $base"
#		echo "Max policy size          : $maxpolicysize"

                echo -n "Apply settings Y/N [Y] : "
                read opt

                [ "$opt" != "n" ] && break

        done

	return 0
}

###############################################################
function verifySuppliedData() {

	local isOk=1

	echoMsg INFO "Verifying supplied data"

	#Pass over mandatory parameters list and report missing settings
	echoMsg DEBUG "Veridying existance of mandatory params"
	for item in ${MANDATORY_PARAMS_LIST[*]}; do
        	[ -z "$item" ] && echoMsg ERROR "Missing mandatory parameter [$item]" && isOk=0
                break
        done

	[ $isOk -eq 0 ] && return 1
		
	#Verify params format

	echoMsg DEBUG "Verifying params format"
	isPolicyOk $maxpolicysize
        [ $? -eq 1 ] && return 1

	return 0
}

###############################################################
function updateConfigFile() {

	echoMsg INFO "Generating configuration file"

        cat ${CONFIG_PATH}/$TEMPLATE_CONFILE > $CONFILE
        if [ $? -ne 0 ]; then
                echoMsg ERROR "Failed to touch the $CONFILE file ..."
                return 1
        fi

	echo >> $CONFILE
	echo "# Dig DNS queries tool" >> $CONFILE
	echo "dig=\"$DIG_TOOL\"" >> $CONFILE
        echo  >> $CONFILE

	echo "# External IP Address of the A10" >> $CONFILE
	echo "external_ip=\"$external_ip\"" >> $CONFILE
	echo  >> $CONFILE

	echo "# Internal IP Address of the A10" >> $CONFILE
	echo "device=\"$internal_ip\"" >> $CONFILE
	echo >> $CONFILE

	echo "# Directory to store the A10 Class Lists" >> $CONFILE
	echo "base=\"$base\"" >> $CONFILE
	echo >> $CONFILE

	echo "# ThreatSTOP Block List Name" >> $CONFILE
	echo "block_list=\"$block_list\"" >> $CONFILE
	echo >> $CONFILE

	echo "# ThreatSTOP Allow List Name" >> $CONFILE
	echo "allow_list=\"$allow_list\"" >> $CONFILE
	echo >> $CONFILE

	echo "# ThreatSTOP DDOS selected zones" >> $CONFILE
	echo "ddos_zonelist=\"${ddos_zonelist}\"" >> $CONFILE
	echo >> $CONFILE

	echo "# A10 username for SSH access" >> $CONFILE
	echo "username=\"$username\"" >> $CONFILE
	echo >> $CONFILE

	echo "# A10 password for SSH access" >> $CONFILE
	echo "password=\"$password\"" >> $CONFILE
	echo >> $CONFILE

	echo "# Enable password for the A10" >> $CONFILE
	echo "enable_pw=\"$enable_pw\"" >> $CONFILE
	echo  >> $CONFILE

	# NOTE : port is also in the template; therefore,
	# we must adjust, rather than append
	sed -i -r "s/^port=.*/port=\"$port\"/" $CONFILE

	return 0
}

############################################
function verifyDeviceAccess() {

        echoMsg INFO "Verifying network access to A10 device $internal_ip"

        echo quit | telnet $internal_ip 22 2>/dev/null  | grep Connected &>/dev/null
        rc=$?

        if [ $rc -ne 0 ] ; then
                echoMsg ERROR "Connection to A10 device $internal_ip failed..."
                return 1
        fi

        echoMsg INFO "Establishing ssh conection"
        echo
        echo "================================================================================"
        echo "We need to make sure that ssh authorized_keys are properly setup between your"
        echo "machine and your A10 device. Please wait for the ssh password prompt and then"
        echo "login into A10 device. Immediately after, type exit and press enter. This will"
        echo "end the ssh session and bring you back into this setup script."
        echo "Please watch for possible ssh conection failures, in which case you will need"
        echo "to resolve the issue before proceeding with this setup script"
        echo "================================================================================"
        echo

        ssh -l ${username} ${internal_ip}
        local rv=$?

        if [ $rv -ne 0 ]; then
                echoMsg ERROR "Failed in ssh connection to A10 device..."
                return 1
        fi

        echo
        echoMsg INFO "Netowrk acceess OK"
        return 0
}

############################################
function setLogrotation() {

        local sys_log_type=0            #This gives us a syslog type
        local sys_log_up_do=0           #This shows if we need to update sys conf file
        local sys_log_conf_do=0         #This shows if we need to update log conf file
        local rc
        local rv

        echoMsg INFO "Prepare logrotation configuration"

        ############################################################
        #Detecting if we are running syslog-ng
        if sudo service syslog-ng status &>/dev/null && ps -C syslog-ng &>/dev/null; then
                sys_log_type=1
                echoMsg INFO "Syslog-ng service has been detected"
                [ ! -f "${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE}" ] && sys_log_up_do=1
        fi

        #Detecting if we are running rsyslog
        if sudo service rsyslog status &>/dev/null && ps -C rsyslogd &>/dev/null; then
                sys_log_type=2
                echoMsg INFO "rsyslog service has been detected"
                [ ! -f "${SYSLOGR_CONF_SYSTEM_PATH}/${SYSLOGNGR_CONF_FILE}" ] && sys_log_up_do=1
        fi

        #Detecting if we are running sysklog
        if sudo service syslog status &>/dev/null && ps -C syslogd &>/dev/null; then
                sys_log_type=3
                echoMsg INFO "syslog (sysklog) service has been detected"
                [ ! -f "${SYSLOGR_CONF_SYSTEM_PATH}/${SYSLOGNGR_CONF_FILE}" ] && sys_log_up_do=1
        fi

        #Running syslog service is not supported 
        if [ $sys_log_type -eq 0 ]; then
                echoMsg ERROR "Unsupported syslog service detected..."
                return 1
        fi

        #Detecting if there is a previous version of logrotation file exists
        if [ ! -f "${SYSLOG_ROTATE_SYSTEM_FILE}" ]; then
                echoMsg DEBUG "No previous logrotation config file has been detected"
                sys_log_conf_do=1
        fi

        if [ $sys_log_type -eq 1 ]; then
                echoMsg DEBUG "Updating ${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} file"
                #echo "sudo su -c cat ${TEMPLATES_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} > ${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE}"
                rc=$(sudo su -c "cat ${TEMPLATES_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} > ${SYSLOGNG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGNG_CONF_FILE} 2>&1")
                rv=$?
        fi

        if [ $sys_log_type -eq 2 ]; then

                local TFILE=${TEMPLATES_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE}                # template
                local RFILE=${SYSLOGR_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE}      # actual
                rv=0
                if [  -z "${internal_ip}" ]; then

                        # will not configure logging without a
                        # device address to lock it down against
                        sudo rm -f ${RFILE}

                else

                        if [ -e "${RFILE}" ]; then
                                echoMsg INFO "Detected existing ${SYSLOGR_CONF_SYSTEM_PATH}/${TEMPLATE_SYSLOGR_CONF_FILE}"
                        else
                                echoMsg DEBUG "Creating ${RFILE}"
                                rc=$(sudo su -c "cat ${TFILE} > ${RFILE} 2>&1")
                        fi

                        # Provides TCP syslog reception
                        #$ModLoad imtcp
                        #$InputTCPServerRun 514
                        # Provides UDP syslog reception
                        #$ModLoad imudp
                        #$UDPServerRun 514
                        RCONF=/etc/rsyslog.conf
                        grep -q -E '^\$ModLoad imtcp$' ${RCONF} || sudo sh -c "echo '\$ModLoad imtcp' >> ${RCONF}"
                        grep -q -E '^\$InputTCPServerRun 514$' ${RCONF} || sudo sh -c "echo '\$InputTCPServerRun 514' >> ${RCONF}"
                        grep -q -E '^\$ModLoad imudp$' ${RCONF} || sudo sh -c "echo '\$ModLoad imudp' >> ${RCONF}"
                        grep -q -E '^\$UDPServerRun 514$' ${RCONF} || sudo sh -c "echo '\$UDPServerRun 514' >> ${RCONF}"

                        # open up firewalld to allow tcp/udp 514; we'll
                        # take advantage of the rsyslog.xml service file
                        # that we provide in this installation
                        if sudo [ -e "${FIREWALLD_RSYSLOG_SERVICE_FILE}" ]; then
                                # from man page for firewall-cmd :
                                # "If you want to make a change in runtime and permanent configuration,
                                #  use the same call with and without the--permanent option."
                                # Also, to have firewalld even know about the rsyslog service we installed
                                # we have to first "reload
                                sudo /bin/firewall-cmd --quiet --reload
                                sudo /bin/firewall-cmd --quiet --reload --permanent
                                sudo /bin/firewall-cmd --quiet --zone=public --add-service=rsyslog
                                sudo /bin/firewall-cmd --quiet --zone=public --add-service=rsyslog --permanent
                                if [ $? -eq 0 ]; then
                                        echoMsg INFO "firewalld opened for rsyslog"
                                else
                                        echoMsg ERROR "couldn't add rsyslog service to the public firewalld zone"
                                fi
                        fi

                        # threatstop logging, restricted to logs from device
                        #:fromhost-ip, isequal, "1.2.3.4" /dev/null
                        sudo sed -r -i "s#^(:fromhost-ip, isequal,).*\$#\1 \"${internal_ip}\" ${TARGET_LOG}#;" ${RFILE}
                        rv=$?
                        if [ $rv -ne 0 ]; then
                                echoMsg ERROR "Error configuring rsyslog ($RFILE)"
                        else
                                echoMsg INFO "rsyslog configured"
                                sudo touch ${TARGET_LOG}
                                sudo service rsyslog restart
                        fi
                fi
        fi

        if [ $sys_log_type -eq 3 ]; then
		rv=0
		grep -q -i threatstop "${SYSKLOG_CONF_SYSTEM_PATH}/${SYSTEM_SYSKLOG_CONF_FILE}"
		if [ $? -ne 0 ]; then
                	echoMsg DEBUG "Updating ${SYSKLOG_CONF_SYSTEM_PATH}/${TEMPLATE_SYSKLOG_CONF_FILE} file"
			# save first /etc/syslog we encounter
			if [ ! -e "${SYSKLOG_CONF_SYSTEM_PATH}/${SYSTEM_SYSKLOG_CONF_FILE}.pre-ts" ]; then
				sudo su -c "cat ${SYSKLOG_CONF_SYSTEM_PATH}/${SYSTEM_SYSKLOG_CONF_FILE} > ${SYSKLOG_CONF_SYSTEM_PATH}/${SYSTEM_SYSKLOG_CONF_FILE}.pre-ts 2>&1"
			fi
                	rc=$(sudo su -c "cat ${TEMPLATES_PATH}/${TEMPLATE_SYSKLOG_CONF_FILE} >> ${SYSKLOG_CONF_SYSTEM_PATH}/${SYSTEM_SYSKLOG_CONF_FILE} 2>&1")
                	rv=$?

			# enable network logging, if not already
			SYSLOGCONF=/etc/sysconfig/syslog
			if [ ! $(grep -q -E '^SYSLOGD_OPTIONS="-r -x -m 0"' ${SYSLOGCONF}) ]; then
				sudo sed -r -i.pre-ts 's/^SYSLOGD_OPTIONS=.*$/SYSLOGD_OPTIONS="-r -x -m 0"/;' ${SYSLOGCONF}
				sudo touch /var/log/remotes.log
				sudo service syslog restart
			fi
		fi
        fi

        [ $rv -gt 0 ] && echoMsg WARN "Failed in syslog configuration files update"

        echoMsg DEBUG "Updating ${SYSLOG_ROTATE_SYSTEM_FILE} file"
        #echo "sudo su -c cat ${TEMPLATES_PATH}/${TEMPLATE_SYSLOG_ROTATE_FILE} > ${SYSLOG_ROTATE_SYSTEM_FILE}"
        rc=$(sudo su -c "cat ${TEMPLATES_PATH}/${TEMPLATE_SYSLOG_ROTATE_FILE} > ${SYSLOG_ROTATE_SYSTEM_FILE} 2>&1")
        rv=$?

        [ $rv -gt 0 ] && echoMsg WARN "Failed in syslog configuration files update"

        return 0
}

############################################
function setCrons() {

        echoMsg INFO "Setting up cron entries for automatic execution"

        local min=$(date +%M)
        local rc
        local rv
        local isError=0

        #For each non exisitng cron install it

	# threatstop user crontab (list updates)
        $CRONTAB -l 2>/dev/null > $TMP_CRONFILE
        if ! cat $TMP_CRONFILE | grep -E "\*/2 \* \* \*" &>/dev/null; then
                #No cron entry for update script job add from template
                cat ${TEMPLATES_PATH}/${SYSLOG_CRON_UPDATE_FILE} >> $TMP_CRONFILE
        fi

        #Update time settings and upload cron
        sed -i "s#^[0-9]\+\(.* \* \* \*\) .*#$min \1 ${SCRIPTS_PATH}/${UPLOADFILE}#" $TMP_CRONFILE
        rc=$($CRONTAB $TMP_CRONFILE 2>&1)
        rv=$?

        if [ $rv -ne 0 ]; then
                echoMsg ERROR "Failed in creating update cron job"
                isError=1
        fi

        #See if we have errors
        [ $isError == 1 ] &&  return 1

	# Generate root crontab (log rotation)
	sudo $CRONTAB -l 2>/dev/null > $TMP_CRONFILE
        if ! cat $TMP_CRONFILE | grep -E "perl.*logrotate.*remotes.*" &>/dev/null; then
                #No cron entry for rotate logs job add from template file
                cat ${TEMPLATES_PATH}/${SYSLOG_CRON_ROTATE_FILE} >> $TMP_CRONFILE
        fi

        #Update time settings and upload cron
        #sudo sed -i "s#^[0-9]\+\(.*perl -e.*remotes\)#$min \1#" $TMP_CRONFILE
        rc=$(sudo $CRONTAB $TMP_CRONFILE 2>&1)
        rv=$?

        if [ $rv -ne 0 ]; then
                echoMsg ERROR "Failed in creating rotate cron job - [$rc]"
                isError=1
        fi

        #See if we have errors
        [ $isError == 1 ] &&  return 1

        return 0
}

############################################
function uploadRules() {

        echoMsg INFO "Running upload rules script for first time... "
        echo
        echo "=============================================================================="
        echo "We are ready to run class list generation for the first time. The script"
        echo "execution output will be saved into $LOGFILE log file."
        echo "NOTE : this will take a few minutes."
        echo "=============================================================================="
        echo

	# Create directory to hold lists
	if [ ! -e "$base" ]; then
		mkdir $base 2>/dev/null
	fi
	if [ ! -w "$base" ]; then
		echo "$base is not writeable"
		return 1
	fi

        echoMsg INFO "Building allow/block lists..."

        ./$UPLOADFILE | tee $LOGFILE
        local rc=$?

        if [ $rc -ne 0 ]; then
                echoMsg ALERT "Failed in class list generation execution"
                echo
                echo "=============================================================================="
                echo "List generation execution failure. The full script execution output is "
                echo "stored in $LOGFILE." 
                echo "Please contact support for assistance, support@threatstop.con"
                echo "=============================================================================="
                echo
                return 1
        fi
        echo

        echoMsg INFO "Class list generation process completed... "

        return 0
}

###############################################################
####################### MAIN PART #############################
###############################################################

WHOAMI=`/usr/bin/whoami 2>/dev/null`
if [ "$WHOAMI" != "threatstop" ]; then
        echo "Please run the setup command as the threatstop user"
        exit 1
fi

/usr/bin/clear
echo 

showPrepMsg

echoMsg INFO "Starting setup.sh script v.$VERSION execution"
#setLock
#[ $? -ne 0 ] && aborting

verifySystemCfg				# See if we have all necessary mandatory scripts and perl modules
[ $? -ne 0 ] && aborting

loadPrevSettings			# Load previous stored settings from config file
[ $? -ne 0 ] && aborting

readCLIParams	$*			# read CLI supplied parameters 
[ $? -ne 0 ] && aborting

readUserData				# ask user for missing data 
[ $? -ne 0 ] && aborting

verifySuppliedData			# verify that we have all mandatory parameters before procceding
[ $? -ne 0 ] && aborting

updateConfigFile			# update config file with supplied settings
[ $? -ne 0 ] && aborting

#verifyDeviceAccess			# Check if we can access remote A10 device
#[ $? -ne 0 ] && aborting

setLogrotation				# Enable log rotation rollback on abort
[ $? -ne 0 ] && rollBackSysConfig && aborting

setCrons				# Enable cron job rollback on abort
[ $? -ne 0 ] && rollBackSysConfig && aborting

uploadRules				# Upload rules for the first time
[ $? -ne 0 ] && rollBackSysConfig && aborting

unLock
echoMsg INFO "Finished setup.sh v.$VERSION execution"

