#!/usr/bin/perl

# /**
#  *
#  * Copyright(c) 2006-2016 ThreatSTOP, Inc. All Rights Reserved
#  *
#  * NOTICE: All information contained herein is, and remains the property
#  * of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and 
#  * technical concepts contained herein are proprietary to ThreatSTOP, Inc. 
#  * and its suppliers and certain aspects thereof are protected by United 
#  * States Patent No. 8,533,822 and United States Patent No. 8,869,237, 
#  * and are protected by trade secret or copyright law.
#  *
#  * US Government Entities: The ThreatSTOP service, software and
#  * documentation, as applicable, are "commercial computer software" and
#  * "commercial computer software documentation" developed exclusively at
#  * private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
#  * 12.212 or DFARS 227 7202 and their successors, as applicable, use,
#  * reproduction and disclosure of the software and documentation is
#  * governed by the terms of Threatstop's commercial agreement.
#  */

use warnings;
use strict;

use Fcntl qw(:flock);
use Data::Dumper;

use FindBin;
use File::Copy;
use lib "$FindBin::Bin/lib/";
use ThreatStop;
use POSIX;

######################################################################
####################### GLOBAL VARIABLES PART ########################
######################################################################

$LOG_LEVEL  = 5; # from ThreatStop.pm
my $LOGGING    = 1;
my $AUTOUP     = 1;  # set to 0 if you do not want for threatstop.conf file being overwritten

my $DATA        = {};
my $PATH_MY    = substr($0,0,rindex($0,'/')+1);
my $T          = gmtime;
my $NOW_STRING = localtime;
my $DIG        = 'dig +tcp -t ';
my $CONFIG_PATH = "/etc/threatstop";
my $CONFILE   = "$CONFIG_PATH/threatstop.conf";
chomp (my $VERSION = `cat $CONFIG_PATH/VERSION`);
my $SSH;
my $MAXPOLICYSIZE = 5000;
my $POLICYCNTR = 0;
my $PAGER;

my $START_EPOCH = time();
my $END_EPOCH;
my $ELAPSED_TIME;

#my $DEBUG=($ARGV[0] && $ARGV[0] eq 'DEBUG')?1:0 ;

####################################################
# data structures to hold all data bound for the class files
# the strategy will be to load these up, then write them out
#
# NOTE : the reason for breaking up the CIDRs into
# /32 and other-than-/32 is for guarding the maximum
# counts per-file that ACOS allows.  We believe each class
# list to accomodate either
# 8,000,000 hosts per list (/32)
# or
# 64,000 nets (other-than-/32)
#
# We'll guard those counts as we write the class lists,
# creating as many class lists as necessary (up to max
# class lists of 255)
#
# Entries from all targets (DDOS and Standard) are merged into a hash
# to remove duplicates
# Duplicates happen across DDOS targets as well as across DDOS and Standard 
# %ALLOWS = { hosts => { Allowed IPs }, nets => { Allowed subnets } }
# %BLOCKS = { hosts => { Blocked IPs }, nets => { Blocked subnets } }
my %ALLOWS = ( hosts => {}, nets => {} );
my %BLOCKS = ( hosts => {}, nets => {} );

$LOG_LEVEL = 6 if ( defined($ARGV[0]) && $ARGV[0] eq "DEBUG");

######################################################################
######################## FUNCTIONS PART ##############################
######################################################################

##################################################
# Remove duplicates from an array
sub uniq {
    my %seen = ();
    my @r = ();
    foreach my $a (@_) {
        unless ($seen{$a}) {
            push @r, $a;
            $seen{$a} = 1;
        }
    }
    return @r;
}

##########################################################
#loadPrevConfData : Running from lib/Threatstop.pm module#
##########################################################
sub loadPrevConfData() {

        my $data_hash = {};
        my $err_msg;

        &printMsg("INFO","Previous configuration found ... Loading config data...");
        if ( &loadConfFileData($CONFILE,$data_hash, \$err_msg) != 0 ) {
                &printMsg("ERROR","No configuration file had been found ...");
                return 1;
        }

        #Now load retrived hash into our data holding hash
        foreach my $key ( keys %{$data_hash}) {
                $DATA->{$key} = $data_hash->{$key};
        }

        if ( ! defined $DATA->{'dns_server'} ) {
                &printMsg("WARN","Missing dns_server parameter ... using default values...");
                $dns_server           = $DNS_LIST;
                $DATA->{'dns_server'} = $dns_server;
        } else {
                $dns_server = $DATA->{'dns_server'};
        }

        if ( ! defined $DATA->{'dig'} ) {
                $DATA->{'dig'} = '/usr/sbin/dig';
        }

        if ( ! defined $DATA->{'port'} ) {
                $DATA->{'port'} = 53;
        }

        return 0;
}

#############################################################
#updateDNSrLists: Running from lib/Threatstop.pm module     #
#############################################################
sub updateDNSLists() {

        #I am not sure why we do this update because this way we can reload
        #not responding DNS servers we clear up in he previous function  
        #but folow the original code flow

	my $dns_new_lst;
        my $dns_new_port;
        my $dns_server = $DATA->{'dns_server'};
        my $dig_tool   = $DATA->{'dig'};
        my $dns_port   = $DATA->{'port'};
        my $conf       = $CONFILE;

        &printMsg("DEBUG","Ready to update DNS lists with params [$dig_tool, $dns_server, $dns_port]");
	return 1 if ( &updateDNSlist("$dig_tool","$dns_server",$dns_port,\$dns_new_lst, \$dns_new_port) != 0 );

	#Update DNS settings in DATA hash
        $DATA->{'dns_server'} = $$dns_new_lst  if ( defined $$dns_new_lst );
        $DATA->{'port'}       = $$dns_new_port if ( defined $$dns_new_port);
        return 0;
}

################################################
#the DNS lookups to get the block lists
sub getlist {

        my $list     = shift;
        my $listtype = shift;
	my $ans_ref  = shift;
	my $dns      = (split(' ',$DATA->{'dns_server'}))[0];

	&printMsg("DEBUG","$DIG ptr -p $DATA->{'port'} \@$dns $list");

        my $cmd      = "$DIG ptr -p $DATA->{'port'} \@$dns $list";
        my ( %answers_net, %answers_ip, @lists );
        my $rc = qx($cmd);

	if (!defined $rc) {
		&printMsg("ERROR","Could not run dig. Likely running out of memory.");
		return 1;
	}

	unless ( $rc =~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s) {
		&printMsg("ERROR","Failed in initial DIG");
		return 1;
	}

        &printMsg("INFO","Getting $listtype list $list");
        while ( $rc =~ /$list\.\s+\d+\s+IN\s+PTR\s+(\S+)\.\s/sg) {
                push @lists, $1;
                &printMsg("DEBUG","$1");
        }

        &printMsg("INFO","Received ", scalar(@lists), " ", ${listtype}, " lists to process");

        for my $l (@lists) {
		&printMsg("DEBUG","$DIG apl -p $DATA->{'port'} \@$dns $l");
                my $cmd = "$DIG apl -p $DATA->{'port'} \@$dns $l";
                $rc = qx($cmd);

		#Remove possible bad format 0.0.0.0/0
		$rc =~ s/0\.0\.0\.0\.0\/0//sg;

		unless ( $rc =~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER:\s+(\d+)/s ) {
			&printMsg("ERROR","DIG $l failed returning error [$rc]");
			return 1;
		}

                &printMsg("DEBUG","$l $1 records");
                while ( $rc =~ /$l\.\s+\d+\s+IN\s+APL\s+\d:(\d+\.\d+\.\d+\.\d+\/\d+)/sg) {
			my $entry = $1;
			$entry =~ m!(\d+\.\d+\.\d+\.\d+)\/(\d+)!;
			$POLICYCNTR++;
			if ($2 == 32) {
				$ans_ref->{hosts}->{"$1 /$2"}++;
			} else {
				$ans_ref->{nets}->{"$1 /$2"}++;
			}
                }
        }
        &printMsg("INFO","Received $POLICYCNTR $listtype results");
        return 0;
}

################################################
# the DNS lookups to get the ddos zone & break into block & allow
# the passed array references have members added to them
#
# return 0 for success; 1 otherwise
sub getddoslist {
	my $types = shift;
	my $allowref = shift;
	my $blockref = shift;
	my $dns      = $DATA->{'dns_server_rpz'};

	# allow 
	my $acount = 0;

	# block
	my $bcount = 0;

	foreach my $type ( sort split(/\s+/, $types) ) {

		chomp $type;
	 	my $tmpzone  = qq($DATA->{'base'}/tmp/$type.axfr);
		my $zone = "ddos-$type.rpz.a10.local";

		my $cmd = "$DIG axfr -p $DATA->{'port'} -y 'threatstop-a10:t0MvBKvxtt7kgWMq+TDDsg==' \@$dns $zone";
		&printMsg("DEBUG", $cmd);

		my $rc = qx($cmd > $tmpzone);
		if (!defined $rc) {
			&printMsg("ERROR","Could not run dig. Likely running out of memory.");
			return 1;
		}

		system("/usr/bin/tail $tmpzone | grep -q -E ';; XFR size: [0-9]+ records'");
		if ( $? != 0 ) {

			&printMsg("ERROR", "Zone transfer failed : $cmd [$?]");
			return 1;
		}

		if ( ! open(ZONE, "<$tmpzone") ) {
			&printMsg("ERROR", "$tmpzone : $!");
			return 1;
		}

		while (<ZONE>) {

			# 32.42.4.18.103.rpz-ip.Alpo.alanin01.rpz.threatstop.local. 14400 IN CNAME rpz-drop.
			if ( /^(\d+)\.(\d+)\.(\d+)\.(\d+)\.(\d+)\.([\w-]+)\.([\w.-]+?)\.\s+(\d+)\s+[\w-]+\s+CNAME\s+([\w-]*?)/ ) {

				my $mask = $1;
				my $ip = "$5.$4.$3.$2";
				my $type =  $6;            # RPZ type ...expected to be rpz-ip
				my $zone = $7;
				my $ttl = $8;
				my $action = $9;

				if ( $type ne 'rpz-ip' ) {
					&printMsg("ERROR", "Encountered RPZ line with type '$type'");
					next;
				}

				if ( $action eq 'rpz-passthru' ) {
					if ( $mask == 32 ) {
						$allowref->{hosts}->{"$ip /$mask"}++;
					} else {
						$allowref->{nets}->{"$ip /$mask"}++;
					}
					$acount++;
				} else {
					if ( $mask == 32 ) {
						$blockref->{hosts}->{"$ip /$mask"}++;
					} else {
						$blockref->{nets}->{"$ip /$mask"}++;
					}
					$bcount++;
				}
			} else {
				&printMsg("DEBUG","skipping $_");
			}
		}
		close ZONE;

		if ($LOG_LEVEL != 6) {
			# Keep temp files while debugging
			unlink $tmpzone or &printMsg("ERROR","Could not remove $tmpzone");
		}
	}

        &printMsg("INFO","Received $acount allow entries from DDOS");
        &printMsg("INFO","Received $bcount block entries from DDOS");

	return 0;
}

################################################
sub writeLists {

	my $refref = shift;   # array of hashes
	my $base = shift;
	my $maxlines = shift; # per file
	my $countref = shift; # next file number to use...we bump it along here
	my $listtype = shift;
	my $linecount = 0;
	my $file_with_space_orig = 'unset';
	my $file_no_space = 'unset';
	my $file_with_space = 'unset';
	my $suffix;
	my $newlist = 1; # triggers list info to be written
	my $listsize = 0;

	foreach my $ref ( @$refref ) {

		if (! defined $ref) {
			# Undefined list; should never be reached
			&printMsg("WARN", "Unexpected list state (contact support)");
			next;
		}

		if (scalar keys %$ref == 0) {
			&printMsg("WARN", "Found empty list");
		}

		$newlist = 1;
		$listsize = scalar keys %$ref;

		foreach my $cidr ( keys %$ref ) {

			# Enable this to see duplicates; it can generate a lot of entries
			#if ($ref->{$cidr} > 1) {
			#	&printMsg("DEBUG", "Removed duplicate entry $cidr\n");
			#}

			if ( $linecount == 0 ) {

				# only define these when they're about to be used
				$suffix = sprintf "%03d", $$countref;
				$file_with_space_orig = "$DATA->{'base'}/tmp/${base}${suffix}.txt";
				$file_no_space =      "$DATA->{'base'}/tmp/${base}${suffix}-nsp.txt";
				$file_with_space =    "$DATA->{'base'}/tmp/${base}${suffix}-wsp.txt";

				&printMsg("DEBUG","Opening $file_with_space_orig");
				&printMsg("DEBUG","Opening $file_no_space");
				&printMsg("DEBUG","Opening $file_with_space");
				if (	! open OUTO, ">$file_with_space_orig" or
					! open OUTN, ">$file_no_space" or
					! open OUTW, ">$file_with_space"
					 ) {

					&printMsg("DEBUG","$!");
					next;
				}

				# file header (mandatory?)
				print OUTO "; Class list - ThreatSTOP-$file_with_space_orig\n";
				print OUTO "; IPv4\n\n";
				print OUTN "; Class list - ThreatSTOP-$file_no_space\n";
				print OUTN "; IPv4\n\n";
				print OUTW "; Class list - ThreatSTOP-$file_with_space\n";
				print OUTW "; IPv4\n\n";
			}

			if ( $newlist ) {

				print OUTO "; $listtype (total: $listsize)\n";
				print OUTN "; $listtype (total: $listsize)\n";
				print OUTW "; $listtype (total: $listsize)\n";
				$newlist = 0;
			}

			my $cidrn = $cidr;
			$cidrn =~ s/\s+//g; # cidr version with no spaces
			print OUTO "$cidr\n";
			print OUTN "$cidrn\n";
			print OUTW "$cidr\n";
			$linecount++;

			if ( $linecount >= $maxlines ) {

				&printMsg("DEBUG","Closing $file_with_space_orig");
				&printMsg("DEBUG","Closing $file_no_space");
				&printMsg("DEBUG","Closing $file_with_space");
				close OUTO;
				close OUTN;
				close OUTW;
				$$countref++;
				$linecount = 0;
			}
		}
	}

	if ( $linecount != 0 && $linecount < $maxlines ) {

		&printMsg("DEBUG","Closing $file_with_space_orig");
		&printMsg("DEBUG","Closing $file_no_space");
		&printMsg("DEBUG","Closing $file_with_space");
		close OUTO;
		close OUTN;
		close OUTW;
		$$countref++;
	}
}

################################################
sub buildLists() {

	my $allowbase = 'allow-';
	my $blockbase = 'block-';
	my $maxlists = 255;
	my $hostmax = 8000000;
	my $hosttype = 'Single IP address';
	my $subnetmax = 64000;
	my $subnettype = 'IP Subnet';
	my $aggregatetype = 'IP/Subnet aggregate';
	my $nomax = LONG_MAX; # from POSIX, needs to be larger than any number of ip/cidrs we'll ever deliver
	my $base = $DATA->{'base'};

	if ( ! -d $base ) {

                &printMsg("ERROR","$base is not a directory");
		return 1;
	}

	system("/bin/mkdir -p $base/tmp");
	if ( $?>>8 != 0 ) {

                &printMsg("ERROR","Could not create $base/tmp");
		return 1;
	}

	&printMsg("INFO","Building allow/deny lists");
	if ( $DATA->{'ddos'} == 1 ) {
		return 1 if ( getddoslist( $DATA->{'ddos_zonelist'}, \%ALLOWS, \%BLOCKS) != 0 );
	}

	# By the time we reach this point, if DDOS is enabled, the size of this process in memory is very lage
	# Unless the VM has plenty of memory, forking will fail
	# Consequently, you can't use system() or qx() from this point on
	# A workaround is to allow the kernel to overcommit memory
	# echo 1 > /proc/sys/vm/overcommit_memory

	return 1 if ( getlist( $DATA->{'allow_list'}, 'allow', \%ALLOWS) != 0 );
	return 1 if ( getlist( $DATA->{'block_list'}, 'block', \%BLOCKS) != 0 );

	my $filecount = 0;
	if ( $DATA->{'aggregate_class_lists'} == 1 ) {

		# GS-878, all blocks in one class lists, all allows in one class list
		$filecount = 0;
		writeLists( [ $ALLOWS{hosts}, $ALLOWS{nets} ], $allowbase, $nomax, \$filecount, $aggregatetype );
		$filecount = 0;
		writeLists( [ $BLOCKS{hosts}, $BLOCKS{nets} ], $blockbase, $nomax, \$filecount, $aggregatetype );
	} else {

		# write class lists
		$filecount = 0;
		writeLists( [ $ALLOWS{hosts} ], $allowbase, $hostmax, \$filecount, $hosttype );
		writeLists( [ $ALLOWS{nets} ], $allowbase, $subnetmax, \$filecount, $subnettype );
		$filecount = 0;
		writeLists( [ $BLOCKS{hosts} ], $blockbase, $hostmax, \$filecount, $hosttype );
		writeLists( [ $BLOCKS{nets} ], $blockbase, $subnetmax, \$filecount, $subnettype );
	}

	if ( $filecount >= $maxlists ) {

		# we shouldn't hit this, but just in case, we'll note it
                &printMsg("ERROR","Created more lists ($filecount) than ACOS can support");
	}

	# the (semi)atomic placement of the new files
	# Needs to be atomic to prevent the client from picking up files while we're updating them
	unlink glob("$base/allow*.txt $base/block*.txt");
	for my $file (glob "$base/tmp/*.txt") {
		move($file, "$base/");
		if ( $? != 0 ) {
			&printMsg("ERROR","Could not remove old files from $base/ and place new files in $base/tmp/");
		}
	}

	return 0;
}

################################################
sub buildListsOld() {

	# XXX placed here to make the globals here happy
	my @ALLOWS = ();
	my @BLOCKS = ();
	my $DDOS_OPTIMIZE = 0; # change to 1 to use Net::CIDR::Lite on the DDOS CIDRs

	&printMsg("INFO","Building allow/deny lists");
	if ( $DATA->{'DDOS'} == 1 ) {
		return 1 if ( getddoslist( $DATA->{'ddos_zonelist'}, \%ALLOWS, \%BLOCKS, $DDOS_OPTIMIZE) != 0 );
	}
	return 1 if ( getlist( $DATA->{'allow_list'}, 'allow', \%ALLOWS) != 0 );
	return 1 if ( getlist( $DATA->{'block_list'}, 'block', \%BLOCKS) != 0 );

	$DATA->{'allow_list'}=~/(.*)(?:-001-neta)|\./;
	my $filename = $1;
	if (!defined $filename) {
	        $DATA->{'allow_list'}=~/(.*)(?:.threatstop.local)|\./;
		$filename = $1;
	}
	open (FILE, ">$DATA->{'base'}/$filename-allow.txt");
	print FILE <<EOHDR ;
; Class list - ThreatSTOP-$filename-allow
; IPv4

EOHDR
# removed from above, per Sunil request (email 6/3/15)
#; class-list-type ipv4
	print FILE "; Single IP address (Total: ".(scalar @{$ALLOWS[0]}).")\n";
	for (@{$ALLOWS[0]}) {
		print FILE "$_ /32\n";
	}
	print FILE "; IP subnet (Total: ".(scalar @{$ALLOWS[1]}).")\n";
	for (@{$ALLOWS[1]}) {
		print FILE "$_\n";
	}
	close FILE;
	$DATA->{'block_list'}=~/(.*)(?:-001-netb)|\./;
	$filename = $1;
	if (!defined $filename) {
	        $DATA->{'block_list'}=~/(.*)(?:.threatstop.local)|\./;
		$filename = $1;
	}
	open (FILE, ">$DATA->{'base'}/$filename-block.txt");
	print FILE <<EOHDR ;
; Class list - ThreatSTOP-$filename-block
; IPv4

EOHDR
# removed from above, per Sunil request (email 6/3/15)
#; class-list-type ipv4
	print FILE "; Single IP address (Total: ".(scalar @{$BLOCKS[0]}).")\n";
	for (@{$BLOCKS[0]}) {
		print FILE "$_ /32\n";
	}
	print FILE "; IP subnet (Total: ".(scalar @{$BLOCKS[1]}).")\n";
	for (@{$BLOCKS[1]}) {
		print FILE "$_\n";
	}
	close FILE;
	return 0;
}

######################################################################
########################## MAIN PART #################################
######################################################################

exit 1 if ( &logInit()                 != 0 );

&printMsg("INFO","Starting $0 v$VERSION on $NOW_STRING");
&setLock();

exit 1 if ( &loadPrevConfData()        != 0 );
# removing the following; it blatantly sets port back to 53, even when tests against 5353 work just fine
#exit 1 if ( &updateDNSLists()          != 0 );
#exit 1 if ( &buildListsOld()              != 0 );
exit 1 if ( &buildLists()              != 0 );

&calculateElapsedTime($START_EPOCH,\$ELAPSED_TIME);

&setUnLock();

$NOW_STRING = localtime;
&printMsg("INFO","Finished on $NOW_STRING after $ELAPSED_TIME");

